"""Unit tests for context fetcher functionality."""

import asyncio

import pytest

from src.autodoc_mcp.core.context_fetcher import (
    ConcurrentContextFetcher,
    create_context_fetcher,
)
from src.autodoc_mcp.core.context_formatter import (
    DocumentationContext,
    PackageDocumentation,
)
from src.autodoc_mcp.exceptions import NetworkError


@pytest.fixture
def mock_cache_manager(mocker):
    """Create mock cache manager for tests."""
    cache_manager = mocker.AsyncMock()
    return cache_manager


@pytest.fixture
def mock_dependency_resolver(mocker):
    """Create mock dependency resolver for tests."""
    resolver = mocker.AsyncMock()
    return resolver


@pytest.fixture
def mock_formatter(mocker):
    """Create mock context formatter for tests."""
    formatter = mocker.Mock()
    return formatter


@pytest.fixture
def mock_network_client(mocker):
    """Create mock network client for tests."""
    client = mocker.AsyncMock()
    return client


@pytest.fixture
def context_fetcher(
    mock_cache_manager, mock_dependency_resolver, mock_formatter, mock_network_client
):
    """Create ConcurrentContextFetcher instance with all dependencies mocked."""
    return ConcurrentContextFetcher(
        cache_manager=mock_cache_manager,
        dependency_resolver=mock_dependency_resolver,
        formatter=mock_formatter,
        network_client=mock_network_client,
    )


@pytest.fixture
def sample_package_info():
    """Sample package info for testing."""
    return {
        "name": "requests",
        "version": "2.28.0",
        "summary": "Python HTTP for Humans.",
        "description": "Requests is a simple HTTP library for Python.",
        "author": "Kenneth Reitz",
        "license": "Apache 2.0",
        "home_page": "https://requests.readthedocs.io",
        "project_urls": {
            "Documentation": "https://requests.readthedocs.io",
            "Source": "https://github.com/psf/requests",
        },
        "classifiers": ["Development Status :: 5 - Production/Stable"],
        "keywords": ["http", "requests", "web"],
    }


@pytest.fixture
def sample_primary_docs():
    """Sample primary package documentation."""
    return PackageDocumentation(
        name="requests",
        version="2.28.0",
        relationship="primary",
        summary="Python HTTP for Humans.",
        key_features=["HTTP/1.1", "Connection pooling", "SSL/TLS verification"],
        main_classes=["Session", "Request", "Response"],
        main_functions=["get", "post", "put", "delete"],
        usage_examples="requests.get('https://api.github.com')",
        why_included="Requested package",
        dependency_level=0,
    )


@pytest.fixture
def sample_dependency_docs():
    """Sample dependency package documentation."""
    return PackageDocumentation(
        name="urllib3",
        version="1.26.12",
        relationship="runtime_dependency",
        summary="HTTP library with thread-safe connection pooling.",
        key_features=["Connection pooling", "Thread-safe", "SSL/TLS"],
        main_classes=["PoolManager", "HTTPConnection"],
        main_functions=["request", "urlopen"],
        why_included="Required by requests",
        dependency_level=1,
    )


class TestConcurrentContextFetcherInitialization:
    """Test ConcurrentContextFetcher initialization."""

    def test_init_with_all_dependencies(
        self,
        mock_cache_manager,
        mock_dependency_resolver,
        mock_formatter,
        mock_network_client,
    ):
        """Test initialization with all required dependencies."""
        fetcher = ConcurrentContextFetcher(
            cache_manager=mock_cache_manager,
            dependency_resolver=mock_dependency_resolver,
            formatter=mock_formatter,
            network_client=mock_network_client,
        )

        assert fetcher.cache_manager == mock_cache_manager
        assert fetcher.dependency_resolver == mock_dependency_resolver
        assert fetcher.formatter == mock_formatter
        assert fetcher.network_client == mock_network_client
        assert fetcher.config is not None

    def test_init_without_network_client(
        self, mock_cache_manager, mock_dependency_resolver, mock_formatter
    ):
        """Test initialization without network client (allowed to be None)."""
        fetcher = ConcurrentContextFetcher(
            cache_manager=mock_cache_manager,
            dependency_resolver=mock_dependency_resolver,
            formatter=mock_formatter,
            network_client=None,
        )

        assert fetcher.network_client is None
        assert fetcher.cache_manager == mock_cache_manager
        assert fetcher.dependency_resolver == mock_dependency_resolver
        assert fetcher.formatter == mock_formatter

    def test_config_loading(self, context_fetcher, mocker):
        """Test that configuration is loaded during initialization."""
        mock_get_config = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.get_config"
        )
        mock_config = mocker.Mock()
        mock_get_config.return_value = mock_config

        # Create new fetcher to trigger config loading
        ConcurrentContextFetcher(
            cache_manager=mocker.AsyncMock(),
            dependency_resolver=mocker.AsyncMock(),
            formatter=mocker.Mock(),
        )

        mock_get_config.assert_called_once()


class TestFetchPackageContextPrimaryOnly:
    """Test fetch_package_context with primary package only."""

    @pytest.mark.asyncio
    async def test_fetch_primary_only_cache_hit(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test fetching primary package with cache hit."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            package_name="requests",
            version_constraint=">=2.0.0",
            include_dependencies=False,
        )

        # Verify result
        assert isinstance(result, DocumentationContext)
        assert result.primary_package == sample_primary_docs
        assert result.runtime_dependencies == []
        assert result.context_scope == "primary_only"
        assert result.total_packages == 1

        # Verify metrics
        assert metrics["cache_hits"] == 1
        assert metrics["cache_misses"] == 0
        assert metrics["concurrent_fetches"] == 0
        assert metrics["failed_fetches"] == 0
        assert metrics["total_time"] > 0

        # Verify mocks were called correctly
        context_fetcher.cache_manager.resolve_and_cache.assert_called_once_with(
            "requests", ">=2.0.0"
        )
        context_fetcher.formatter.format_primary_package.assert_called_once_with(
            sample_package_info, True
        )

    @pytest.mark.asyncio
    async def test_fetch_primary_only_cache_miss(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test fetching primary package with cache miss."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, False)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            package_name="requests", include_dependencies=False
        )

        # Verify metrics reflect cache miss
        assert metrics["cache_hits"] == 0
        assert metrics["cache_misses"] == 1
        assert result.primary_package == sample_primary_docs

    @pytest.mark.asyncio
    async def test_fetch_primary_with_version_constraint(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test fetching primary package with version constraint."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, False)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )

        # Execute
        await context_fetcher.fetch_package_context(
            package_name="requests",
            version_constraint="~=2.28.0",
            include_dependencies=False,
        )

        # Verify version constraint passed correctly
        context_fetcher.cache_manager.resolve_and_cache.assert_called_once_with(
            "requests", "~=2.28.0"
        )


class TestFetchPackageContextWithDependencies:
    """Test fetch_package_context with dependencies included."""

    @pytest.mark.asyncio
    async def test_fetch_with_runtime_dependencies_success(
        self,
        context_fetcher,
        sample_package_info,
        sample_primary_docs,
        sample_dependency_docs,
        mocker,
    ):
        """Test fetching package with runtime dependencies successfully."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=["urllib3", "charset-normalizer"])
        )

        # Mock concurrent dependency fetching
        mock_fetch_deps = mocker.patch.object(
            context_fetcher,
            "_fetch_dependencies_concurrently",
            return_value=[sample_dependency_docs],
        )

        # Mock config values
        context_fetcher.config.max_dependency_context = 10
        context_fetcher.config.max_context_tokens = 50000

        # Mock context window manager
        mock_context_manager = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.ContextWindowManager"
        )
        mock_context_manager.fit_to_window.return_value = mocker.Mock(
            primary_package=sample_primary_docs,
            runtime_dependencies=[sample_dependency_docs],
            context_scope="with_runtime (1 deps)",
            total_packages=2,
            token_estimate=1500,
            truncated_packages=[],
        )

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            package_name="requests",
            include_dependencies=True,
            context_scope="smart",
        )

        # Verify dependency resolution was called
        context_fetcher.dependency_resolver.resolve_context_dependencies.assert_called_once_with(
            package_name="requests",
            version="2.28.0",
            max_dependencies=10,
            max_tokens=50000,
        )

        # Verify concurrent fetching was called
        mock_fetch_deps.assert_called_once_with(
            ["urllib3", "charset-normalizer"], "requests", metrics
        )

        # Verify result structure
        assert result.total_packages == 2
        assert len(result.runtime_dependencies) == 1
        assert metrics["concurrent_fetches"] == 2

    @pytest.mark.asyncio
    async def test_fetch_with_max_dependencies_limit(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test fetching with max_dependencies parameter."""
        # Reset config to defaults (in case modified by earlier tests)
        context_fetcher.config.max_context_tokens = 30000

        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=[])
        )

        # Execute with custom max_dependencies
        await context_fetcher.fetch_package_context(
            package_name="requests",
            include_dependencies=True,
            max_dependencies=3,
        )

        # Verify max_dependencies was used
        context_fetcher.dependency_resolver.resolve_context_dependencies.assert_called_once_with(
            package_name="requests",
            version="2.28.0",
            max_dependencies=3,
            max_tokens=30000,  # Default from config
        )

    @pytest.mark.asyncio
    async def test_fetch_with_max_tokens_limit(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test fetching with max_tokens parameter."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=[])
        )

        # Mock config to control max_dependency_context
        context_fetcher.config.max_dependency_context = 10

        # Execute with custom max_tokens and non-smart context scope to get fallback
        await context_fetcher.fetch_package_context(
            package_name="requests",
            include_dependencies=True,
            context_scope="runtime",  # Non-smart scope to get fallback of 5
            max_tokens=25000,
        )

        # Verify max_tokens was used and fallback max_dependencies
        context_fetcher.dependency_resolver.resolve_context_dependencies.assert_called_once_with(
            package_name="requests",
            version="2.28.0",
            max_dependencies=5,  # Fallback for non-smart scope
            max_tokens=25000,
        )

    @pytest.mark.asyncio
    async def test_fetch_with_no_dependencies_found(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test fetching when no dependencies are resolved."""
        # Setup mocks - empty dependency list
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=[])
        )

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            package_name="requests", include_dependencies=True
        )

        # Verify result contains only primary package
        assert result.total_packages == 1
        assert result.runtime_dependencies == []
        assert result.context_scope == "primary_only"
        assert metrics["concurrent_fetches"] == 0


class TestFetchPackageContextScoping:
    """Test different context scoping options."""

    @pytest.mark.asyncio
    async def test_context_scope_primary_only(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test context_scope='primary_only' ignores dependencies."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )

        # Execute with primary_only scope
        result, metrics = await context_fetcher.fetch_package_context(
            package_name="requests",
            include_dependencies=True,
            context_scope="primary_only",
        )

        # Dependencies should still be resolved with primary_only scope
        # The behavior depends on the implementation of dependency resolver
        assert result.primary_package == sample_primary_docs

    @pytest.mark.asyncio
    async def test_context_scope_runtime(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test context_scope='runtime' includes runtime dependencies."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=[])
        )

        # Execute with runtime scope
        await context_fetcher.fetch_package_context(
            package_name="requests",
            include_dependencies=True,
            context_scope="runtime",
        )

        # Verify dependency resolution was called with runtime scope
        context_fetcher.dependency_resolver.resolve_context_dependencies.assert_called_once()

    @pytest.mark.asyncio
    async def test_context_scope_smart_uses_config_limits(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test context_scope='smart' uses configuration limits."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=[])
        )

        # Mock config
        context_fetcher.config.max_dependency_context = 15
        context_fetcher.config.max_context_tokens = 75000

        # Execute with smart scope
        await context_fetcher.fetch_package_context(
            package_name="requests",
            include_dependencies=True,
            context_scope="smart",
        )

        # Verify config limits were used
        context_fetcher.dependency_resolver.resolve_context_dependencies.assert_called_once_with(
            package_name="requests",
            version="2.28.0",
            max_dependencies=15,
            max_tokens=75000,
        )


class TestConcurrentDependencyFetching:
    """Test concurrent dependency fetching functionality."""

    @pytest.mark.asyncio
    async def test_fetch_dependencies_concurrently_success(
        self, context_fetcher, sample_dependency_docs, mocker
    ):
        """Test successful concurrent dependency fetching."""
        # Mock single dependency fetch
        mock_fetch_single = mocker.patch.object(
            context_fetcher,
            "_fetch_single_dependency_with_semaphore",
            return_value=sample_dependency_docs,
        )

        dependency_names = ["urllib3", "charset-normalizer", "idna"]
        performance_metrics = {
            "cache_hits": 0,
            "cache_misses": 0,
            "failed_fetches": 0,
        }

        # Execute
        results = await context_fetcher._fetch_dependencies_concurrently(
            dependency_names, "requests", performance_metrics
        )

        # Verify results
        assert len(results) == 3
        assert all(isinstance(result, PackageDocumentation) for result in results)
        assert mock_fetch_single.call_count == 3

    @pytest.mark.asyncio
    async def test_fetch_dependencies_concurrently_empty_list(self, context_fetcher):
        """Test concurrent fetching with empty dependency list."""
        results = await context_fetcher._fetch_dependencies_concurrently(
            [], "requests", {}
        )

        assert results == []

    @pytest.mark.asyncio
    async def test_fetch_dependencies_concurrently_timeout(
        self, context_fetcher, mocker
    ):
        """Test concurrent fetching with timeout."""

        # Mock fetch to simulate timeout by raising TimeoutError
        async def timeout_fetch(*args, **kwargs):
            raise TimeoutError("Simulated timeout")

        mocker.patch.object(
            context_fetcher,
            "_fetch_single_dependency_with_semaphore",
            side_effect=timeout_fetch,
        )

        dependency_names = ["slow-package"]
        performance_metrics = {"failed_fetches": 0}

        # Execute - should timeout
        results = await context_fetcher._fetch_dependencies_concurrently(
            dependency_names, "requests", performance_metrics
        )

        # Should return timeout exceptions
        assert len(results) == 1
        assert isinstance(results[0], Exception)
        assert "timeout" in str(results[0]).lower()

    @pytest.mark.asyncio
    async def test_fetch_single_dependency_with_semaphore(
        self, context_fetcher, sample_dependency_docs, mocker
    ):
        """Test semaphore-controlled single dependency fetching."""
        # Mock the actual fetch method
        mock_fetch = mocker.patch.object(
            context_fetcher,
            "_fetch_single_dependency",
            return_value=sample_dependency_docs,
        )

        semaphore = asyncio.Semaphore(1)
        performance_metrics = {}

        # Execute
        result = await context_fetcher._fetch_single_dependency_with_semaphore(
            semaphore, "urllib3", "requests", performance_metrics
        )

        assert result == sample_dependency_docs
        mock_fetch.assert_called_once_with("urllib3", "requests", performance_metrics)

    @pytest.mark.asyncio
    async def test_fetch_single_dependency_success(
        self, context_fetcher, sample_package_info, sample_dependency_docs, mocker
    ):
        """Test fetching single dependency successfully."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_dependency_package.return_value = (
            sample_dependency_docs
        )

        performance_metrics = {"cache_hits": 0, "cache_misses": 0}

        # Execute
        result = await context_fetcher._fetch_single_dependency(
            "urllib3", "requests", performance_metrics
        )

        # Verify result
        assert result == sample_dependency_docs
        assert performance_metrics["cache_hits"] == 1

        # Verify mocks called correctly
        context_fetcher.cache_manager.resolve_and_cache.assert_called_once_with(
            "urllib3"
        )
        context_fetcher.formatter.format_dependency_package.assert_called_once_with(
            sample_package_info, "requests"
        )

    @pytest.mark.asyncio
    async def test_fetch_single_dependency_failure(self, context_fetcher, mocker):
        """Test fetching single dependency with failure."""
        # Setup mocks to raise exception
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            side_effect=NetworkError("Network failed")
        )

        performance_metrics = {"cache_hits": 0, "cache_misses": 0}

        # Execute
        result = await context_fetcher._fetch_single_dependency(
            "bad-package", "requests", performance_metrics
        )

        # Should return minimal documentation for failed fetch
        assert isinstance(result, PackageDocumentation)
        assert result.name == "bad-package"
        assert result.version == "unknown"
        assert "fetch failed" in result.summary
        assert "Required by requests" in result.why_included


class TestErrorHandling:
    """Test error handling in context fetching."""

    @pytest.mark.asyncio
    async def test_fetch_package_context_primary_fetch_error(
        self, context_fetcher, mocker
    ):
        """Test error handling when primary package fetch fails."""
        # Setup mock to raise exception
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            side_effect=NetworkError("Package not found")
        )

        # Should re-raise the exception
        with pytest.raises(NetworkError, match="Package not found"):
            await context_fetcher.fetch_package_context("nonexistent-package")

    @pytest.mark.asyncio
    async def test_fetch_package_context_dependency_resolution_error(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test error handling when dependency resolution fails."""
        # Setup mocks - primary succeeds, dependency resolution fails
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(side_effect=Exception("Dependency resolution failed"))
        )

        # Mock context window manager
        mock_context_manager = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.ContextWindowManager"
        )
        mock_context_manager.fit_to_window.return_value = mocker.Mock(
            primary_package=sample_primary_docs,
            context_scope="primary_only (deps_failed)",
            token_estimate=500,
        )

        # Should return primary-only context with error indication
        result, metrics = await context_fetcher.fetch_package_context(
            "requests", include_dependencies=True
        )

        # Verify graceful degradation
        assert result.primary_package == sample_primary_docs
        assert "deps_failed" in result.context_scope
        assert metrics["total_time"] > 0

    @pytest.mark.asyncio
    async def test_fetch_package_context_mixed_dependency_results(
        self,
        context_fetcher,
        sample_package_info,
        sample_primary_docs,
        sample_dependency_docs,
        mocker,
    ):
        """Test handling mixed success/failure results in dependency fetching."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=["urllib3", "bad-package"])
        )

        # Mock concurrent fetching with mixed results
        mocker.patch.object(
            context_fetcher,
            "_fetch_dependencies_concurrently",
            return_value=[
                sample_dependency_docs,
                Exception("Failed to fetch bad-package"),
            ],
        )

        # Mock context window manager
        mock_context_manager = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.ContextWindowManager"
        )
        expected_context = mocker.Mock(
            primary_package=sample_primary_docs,
            runtime_dependencies=[sample_dependency_docs],
            context_scope="with_runtime (1 deps)",
            total_packages=2,
            truncated_packages=["Failed: Failed to fetch bad-package"],
            token_estimate=1200,
        )
        mock_context_manager.fit_to_window.return_value = expected_context

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            "requests", include_dependencies=True
        )

        # Verify partial success handling
        assert result == expected_context
        assert metrics["failed_fetches"] == 1
        assert metrics["concurrent_fetches"] == 2

    @pytest.mark.asyncio
    async def test_fetch_package_context_early_primary_failure(
        self, context_fetcher, mocker
    ):
        """Test early failure before primary_docs is created."""
        # Setup mock to fail immediately
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            side_effect=Exception("Immediate failure")
        )

        # Should re-raise the exception since primary_docs doesn't exist
        with pytest.raises(Exception, match="Immediate failure"):
            await context_fetcher.fetch_package_context("bad-package")


class TestPerformanceMetrics:
    """Test performance metrics collection."""

    @pytest.mark.asyncio
    async def test_performance_metrics_structure(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test that all expected performance metrics are collected."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, False)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            "requests", include_dependencies=False
        )

        # Verify all expected metrics are present
        expected_keys = [
            "primary_fetch_time",
            "dependency_resolution_time",
            "dependency_fetch_time",
            "total_time",
            "cache_hits",
            "cache_misses",
            "concurrent_fetches",
            "failed_fetches",
        ]

        for key in expected_keys:
            assert key in metrics
            assert isinstance(metrics[key], int | float)

        # Verify time metrics are reasonable
        assert metrics["total_time"] > 0
        assert metrics["primary_fetch_time"] >= 0
        assert metrics["total_time"] >= metrics["primary_fetch_time"]

    @pytest.mark.asyncio
    async def test_timing_metrics_accuracy(
        self, context_fetcher, sample_package_info, sample_primary_docs, mocker
    ):
        """Test timing metrics accuracy by introducing controlled delays."""
        # Mock time.time to control timing
        # Need extra timestamps for all the time.time() calls in the method
        mock_times = [
            0.0,
            0.1,
            0.2,
            0.2,
            0.2,
            0.2,
        ]  # start, primary_start_end, total_end, extras for all time calls
        mock_time = mocker.patch("src.autodoc_mcp.core.context_fetcher.time.time")
        mock_time.side_effect = mock_times

        # Setup other mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            "requests", include_dependencies=False
        )

        # Verify timing calculations
        assert metrics["primary_fetch_time"] == 0.1  # 0.1 - 0.0
        assert metrics["total_time"] == 0.2  # 0.2 - 0.0


class TestContextWindowManagement:
    """Test context window management integration."""

    @pytest.mark.asyncio
    async def test_context_window_fitting(
        self,
        context_fetcher,
        sample_package_info,
        sample_primary_docs,
        sample_dependency_docs,
        mocker,
    ):
        """Test that context is fitted to token window."""
        # Setup mocks
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(sample_package_info, True)
        )
        context_fetcher.formatter.format_primary_package.return_value = (
            sample_primary_docs
        )
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=["urllib3"])
        )

        # Mock concurrent dependency fetching
        mocker.patch.object(
            context_fetcher,
            "_fetch_dependencies_concurrently",
            return_value=[sample_dependency_docs],
        )

        # Mock context window manager
        mock_context_manager = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.ContextWindowManager"
        )
        fitted_context = mocker.Mock(
            primary_package=sample_primary_docs,
            runtime_dependencies=[sample_dependency_docs],
            truncated_packages=["some-large-package"],
            token_estimate=45000,
        )
        mock_context_manager.fit_to_window.return_value = fitted_context

        # Execute with token limit
        result, metrics = await context_fetcher.fetch_package_context(
            "requests", include_dependencies=True, max_tokens=50000
        )

        # Verify context window manager was called
        mock_context_manager.fit_to_window.assert_called_once()
        args, kwargs = mock_context_manager.fit_to_window.call_args
        assert args[1] == 50000  # max_tokens parameter

        # Verify fitted context is returned
        assert result == fitted_context


class TestCreateContextFetcher:
    """Test factory function for creating context fetcher."""

    @pytest.mark.asyncio
    async def test_create_context_fetcher_with_dependencies(self, mocker):
        """Test create_context_fetcher factory function."""
        # Mock all the components
        mock_cache_manager = mocker.AsyncMock()
        mock_network_client = mocker.AsyncMock()
        mock_network_client.__aenter__ = mocker.AsyncMock(
            return_value=mock_network_client
        )

        mock_network_client_class = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.NetworkResilientClient"
        )
        mock_network_client_class.return_value = mock_network_client

        mock_resolver_class = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.DependencyResolver"
        )
        mock_formatter_class = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.ContextDocumentationFormatter"
        )

        # Execute
        fetcher = await create_context_fetcher(mock_cache_manager)

        # Verify all components were created
        mock_network_client_class.assert_called_once()
        mock_network_client.__aenter__.assert_called_once()
        mock_resolver_class.assert_called_once_with(mock_network_client)
        mock_formatter_class.assert_called_once()

        # Verify fetcher was created properly
        assert isinstance(fetcher, ConcurrentContextFetcher)
        assert fetcher.cache_manager == mock_cache_manager
        assert fetcher.network_client == mock_network_client


class TestIntegrationScenarios:
    """Test realistic integration scenarios."""

    @pytest.mark.asyncio
    async def test_real_world_package_with_many_dependencies(
        self, context_fetcher, mocker
    ):
        """Test realistic scenario with package having many dependencies."""
        # Mock primary package (fastapi)
        primary_info = {
            "name": "fastapi",
            "version": "0.104.1",
            "summary": "FastAPI framework, high performance web API framework",
            "description": "FastAPI is a modern web framework for Python",
        }
        primary_docs = PackageDocumentation(
            name="fastapi",
            version="0.104.1",
            relationship="primary",
            summary="FastAPI framework",
            why_included="Requested package",
            dependency_level=0,
        )

        # Setup mocks for realistic scenario
        context_fetcher.cache_manager.resolve_and_cache = mocker.AsyncMock(
            return_value=(primary_info, False)
        )
        context_fetcher.formatter.format_primary_package.return_value = primary_docs

        # Many dependencies
        dependencies = [
            "pydantic",
            "starlette",
            "typing-extensions",
            "uvicorn",
            "httpx",
        ]
        context_fetcher.dependency_resolver.resolve_context_dependencies = (
            mocker.AsyncMock(return_value=dependencies)
        )

        # Mock successful dependency fetches
        dependency_docs = [
            PackageDocumentation(
                name=dep,
                version="1.0.0",
                relationship="runtime_dependency",
                why_included="Required by fastapi",
                dependency_level=1,
            )
            for dep in dependencies
        ]

        mocker.patch.object(
            context_fetcher,
            "_fetch_dependencies_concurrently",
            return_value=dependency_docs,
        )

        # Mock context window manager
        mock_context_manager = mocker.patch(
            "src.autodoc_mcp.core.context_fetcher.ContextWindowManager"
        )
        fitted_context = mocker.Mock(
            primary_package=primary_docs,
            runtime_dependencies=dependency_docs,
            total_packages=6,
            context_scope="with_runtime (5 deps)",
            token_estimate=35000,
            truncated_packages=[],
        )
        mock_context_manager.fit_to_window.return_value = fitted_context

        # Execute
        result, metrics = await context_fetcher.fetch_package_context(
            "fastapi", include_dependencies=True, context_scope="smart"
        )

        # Verify realistic results
        assert result.total_packages == 6
        assert len(result.runtime_dependencies) == 5
        assert metrics["concurrent_fetches"] == 5
        assert metrics["cache_misses"] >= 1  # At least primary package



"""Configuration management for AutoDocs MCP Server."""

import os
from functools import lru_cache
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from pydantic import BaseModel, Field, field_validator, model_validator

from .security import validate_pypi_url


class AutoDocsConfig(BaseModel):
    """Application configuration with comprehensive validation."""

    cache_dir: Path = Field(default_factory=lambda: Path.home() / ".autodocs" / "cache")
    max_concurrent: int = Field(default=10, ge=1, le=100)
    request_timeout: int = Field(default=30, ge=5, le=300)
    log_level: str = Field(default="INFO")
    pypi_base_url: str = Field(default="https://pypi.org/pypi")

    # Version-based caching settings
    max_cached_versions_per_package: int = Field(default=3, ge=1, le=10)
    cache_cleanup_days: int = Field(default=90, ge=1)

    # Context settings
    max_dependency_context: int = Field(default=8, ge=1, le=20)
    max_context_tokens: int = Field(default=30000, ge=1000, le=100000)

    # Performance and rate limiting settings
    max_retry_attempts: int = Field(default=3, ge=1, le=10)
    base_retry_delay: float = Field(default=1.0, ge=0.1, le=10.0)
    max_retry_delay: float = Field(default=30.0, ge=1.0, le=300.0)
    circuit_breaker_threshold: int = Field(default=5, ge=1, le=50)
    circuit_breaker_timeout: float = Field(default=60.0, ge=10.0, le=3600.0)
    rate_limit_requests_per_minute: int = Field(default=60, ge=1, le=1000)
    max_documentation_size: int = Field(default=50000, ge=1000, le=1000000)

    # Deployment settings
    environment: str = Field(default="development")
    debug_mode: bool = Field(default=False)

    @field_validator("log_level")
    @classmethod
    def validate_log_level(cls, v: str) -> str:
        """Validate log level."""
        valid_levels = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}
        if v.upper() not in valid_levels:
            raise ValueError(f"Invalid log level: {v}. Must be one of {valid_levels}")
        return v.upper()

    @field_validator("pypi_base_url")
    @classmethod
    def validate_pypi_url_enhanced(cls, v: str) -> str:
        """Enhanced PyPI URL validation."""
        try:
            # First use the security module validation
            validated_url = validate_pypi_url(v)

            # Additional security checks
            parsed = urlparse(validated_url)
            if not parsed.scheme or not parsed.netloc:
                raise ValueError("URL must have scheme and netloc")

            # Security check for trusted domains
            trusted_domains = {"pypi.org", "test.pypi.org", "localhost"}
            if parsed.netloc not in trusted_domains and not v.startswith(
                "http://localhost"
            ):
                raise ValueError(f"Untrusted PyPI domain: {parsed.netloc}")

            return validated_url
        except Exception as e:
            raise ValueError(f"Invalid PyPI URL '{v}': {e}") from e

    @field_validator("environment")
    @classmethod
    def validate_environment(cls, v: str) -> str:
        """Validate deployment environment."""
        valid_envs = {"development", "staging", "production"}
        if v not in valid_envs:
            raise ValueError(f"Invalid environment: {v}. Must be one of {valid_envs}")
        return v

    @model_validator(mode="after")
    def validate_retry_delays(self) -> "AutoDocsConfig":
        """Ensure max_retry_delay > base_retry_delay."""
        if self.max_retry_delay <= self.base_retry_delay:
            raise ValueError("max_retry_delay must be greater than base_retry_delay")
        return self

    def model_post_init(self, __context: Any) -> None:
        """Post-initialization validation and setup."""
        # Ensure cache directory exists and is writable
        try:
            self.cache_dir.mkdir(parents=True, exist_ok=True)
            # Test write permissions
            test_file = self.cache_dir / ".write_test"
            test_file.write_text("test")
            test_file.unlink()
        except Exception as e:
            raise ValueError(
                f"Cache directory not writable: {self.cache_dir}. Error: {e}"
            ) from e

    @classmethod
    def from_env(cls) -> "AutoDocsConfig":
        """Load configuration from environment with enhanced validation."""
        try:
            config_data = {}

            # Map environment variables to config fields
            env_mappings = {
                "AUTODOCS_CACHE_DIR": "cache_dir",
                "AUTODOCS_MAX_CONCURRENT": ("max_concurrent", int),
                "AUTODOCS_REQUEST_TIMEOUT": ("request_timeout", int),
                "AUTODOCS_LOG_LEVEL": "log_level",
                "AUTODOCS_PYPI_URL": "pypi_base_url",
                "AUTODOCS_MAX_VERSIONS": ("max_cached_versions_per_package", int),
                "AUTODOCS_CLEANUP_DAYS": ("cache_cleanup_days", int),
                "AUTODOCS_MAX_DEPS": ("max_dependency_context", int),
                "AUTODOCS_MAX_TOKENS": ("max_context_tokens", int),
                "AUTODOCS_MAX_RETRY_ATTEMPTS": ("max_retry_attempts", int),
                "AUTODOCS_BASE_RETRY_DELAY": ("base_retry_delay", float),
                "AUTODOCS_MAX_RETRY_DELAY": ("max_retry_delay", float),
                "AUTODOCS_CIRCUIT_BREAKER_THRESHOLD": (
                    "circuit_breaker_threshold",
                    int,
                ),
                "AUTODOCS_CIRCUIT_BREAKER_TIMEOUT": ("circuit_breaker_timeout", float),
                "AUTODOCS_RATE_LIMIT_RPM": ("rate_limit_requests_per_minute", int),
                "AUTODOCS_MAX_DOC_SIZE": ("max_documentation_size", int),
                "AUTODOCS_ENVIRONMENT": "environment",
                "AUTODOCS_DEBUG": ("debug_mode", lambda x: x.lower() == "true"),
            }

            for env_var, config_key in env_mappings.items():
                value = os.getenv(env_var)
                if value is not None:
                    if isinstance(config_key, tuple):
                        key, converter = config_key
                        config_data[key] = converter(value)
                    else:
                        config_data[config_key] = value

            # Handle Path conversion for cache_dir
            if "cache_dir" in config_data:
                config_data["cache_dir"] = Path(config_data["cache_dir"])

            return cls(**config_data)

        except Exception as e:
            raise ValueError(f"Configuration validation failed: {e}") from e

    def validate_production_readiness(self) -> list[str]:
        """Validate configuration for production deployment."""
        issues = []

        if self.environment == "production":
            if self.debug_mode:
                issues.append("Debug mode should not be enabled in production")

            if not self.pypi_base_url.startswith("https://"):
                issues.append("HTTPS required for production PyPI URL")

            if self.log_level == "DEBUG":
                issues.append("Debug logging not recommended for production")

            if self.max_concurrent > 50:
                issues.append("High concurrency may overwhelm PyPI API")

        return issues


@lru_cache
def get_config() -> AutoDocsConfig:
    """Get cached configuration instance."""
    return AutoDocsConfig.from_env()



"""File-based cache manager for package documentation."""

import contextlib
import json
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any

from structlog import get_logger

from ..exceptions import CacheError
from ..models import CacheEntry, PackageInfo
from ..security import sanitize_cache_key
from .error_formatter import ErrorFormatter, ErrorSeverity, FormattedError

logger = get_logger(__name__)


class CacheManagerInterface(ABC):
    """Interface for cache management."""

    @abstractmethod
    async def get(self, cache_key: str) -> CacheEntry | None:
        """Retrieve cached entry if exists."""

    @abstractmethod
    async def set(self, cache_key: str, package_info: PackageInfo) -> None:
        """Store entry in cache."""

    @abstractmethod
    async def invalidate(self, cache_key: str | None = None) -> None:
        """Invalidate specific key or entire cache."""

    @abstractmethod
    async def initialize(self) -> None:
        """Initialize cache storage."""

    @abstractmethod
    async def resolve_and_cache(
        self, package_name: str, version_constraint: str | None = None
    ) -> tuple[dict[str, Any], bool]:
        """Resolve package version and retrieve from cache or fetch fresh."""


class FileCacheManager(CacheManagerInterface):
    """JSON file-based cache implementation."""

    def __init__(self, cache_dir: Path):
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    async def initialize(self) -> None:
        """Ensure cache directory exists."""
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        logger.info("Cache manager initialized", cache_dir=str(self.cache_dir))

    async def get(self, cache_key: str) -> CacheEntry | None:
        """Retrieve cached entry by version-based key."""
        sanitized_key = sanitize_cache_key(cache_key)
        cache_file = self.cache_dir / f"{sanitized_key}.json"

        if not cache_file.exists():
            logger.debug("Cache miss", cache_key=cache_key)
            return None

        try:
            data = json.loads(cache_file.read_text())

            # Parse the cached data back to models
            package_info = PackageInfo(**data["data"])
            cache_entry = CacheEntry(
                data=package_info,
                timestamp=datetime.fromisoformat(data["timestamp"]),
                version=data["version"],
            )

            logger.debug("Cache hit", cache_key=cache_key, version=cache_entry.version)
            return cache_entry

        except (json.JSONDecodeError, KeyError, ValueError) as e:
            logger.warning("Corrupted cache entry", cache_key=cache_key, error=str(e))
            # Remove corrupted cache file
            with contextlib.suppress(OSError):
                cache_file.unlink()
            return None

    async def set(self, cache_key: str, package_info: PackageInfo) -> None:
        """Store package info in cache with version-based key."""
        sanitized_key = sanitize_cache_key(cache_key)
        cache_file = self.cache_dir / f"{sanitized_key}.json"

        cache_entry = CacheEntry(
            data=package_info,
            timestamp=datetime.now(),
            version=package_info.version,
        )

        try:
            # Serialize for JSON storage
            cache_data = {
                "data": package_info.model_dump(),
                "timestamp": cache_entry.timestamp.isoformat(),
                "version": cache_entry.version,
            }

            cache_file.write_text(json.dumps(cache_data, indent=2))
            logger.info(
                "Cached package info",
                cache_key=cache_key,
                package=package_info.name,
                version=package_info.version,
            )

        except (OSError, TypeError) as e:
            raise CacheError(f"Failed to cache {cache_key}: {e}") from e

    async def invalidate(self, cache_key: str | None = None) -> None:
        """Invalidate specific cache entry or entire cache."""
        if cache_key:
            sanitized_key = sanitize_cache_key(cache_key)
            cache_file = self.cache_dir / f"{sanitized_key}.json"
            try:
                if cache_file.exists():
                    cache_file.unlink()
                    logger.info("Invalidated cache entry", cache_key=cache_key)
            except OSError as e:
                logger.warning(
                    "Failed to invalidate cache entry",
                    cache_key=cache_key,
                    error=str(e),
                )
        else:
            # Clear entire cache
            try:
                for cache_file in self.cache_dir.glob("*.json"):
                    cache_file.unlink()
                logger.info("Cleared entire cache", cache_dir=str(self.cache_dir))
            except OSError as e:
                logger.error("Failed to clear cache", error=str(e))
                raise CacheError(f"Failed to clear cache: {e}") from e

    async def list_cached_packages(self) -> list[str]:
        """List all cached package keys."""
        try:
            return [f.stem for f in self.cache_dir.glob("*.json")]
        except OSError as e:
            logger.error("Failed to list cached packages", error=str(e))
            return []

    async def get_cache_stats(self) -> dict[str, Any]:
        """Get cache statistics."""
        try:
            cache_files = list(self.cache_dir.glob("*.json"))
            total_size = sum(f.stat().st_size for f in cache_files)

            return {
                "total_entries": len(cache_files),
                "total_size_bytes": total_size,
                "cache_dir": str(self.cache_dir),
            }
        except OSError as e:
            logger.error("Failed to get cache stats", error=str(e))
            return {"error": str(e)}

    def get_cached_entry_safe(
        self, cache_key: str
    ) -> tuple[CacheEntry | None, list[FormattedError]]:
        """Get cached entry with error handling."""
        errors: list[FormattedError] = []

        try:
            sanitized_key = sanitize_cache_key(cache_key)
            cache_file = self.cache_dir / f"{sanitized_key}.json"

            if not cache_file.exists():
                return None, errors

            # Check if cache file is corrupted
            try:
                with cache_file.open("r", encoding="utf-8") as f:
                    content = f.read().strip()

                if not content:
                    # Empty cache file
                    cache_file.unlink()  # Remove corrupted file
                    errors.append(
                        FormattedError(
                            message=f"Corrupted cache file removed: {cache_key}",
                            suggestion="The cache file was empty and has been cleaned up automatically.",
                            severity=ErrorSeverity.INFO,
                            error_code="cache_corruption_fixed",
                        )
                    )
                    return None, errors

                data = json.loads(content)

                # Parse the cached data back to models
                package_info = PackageInfo(**data["data"])
                cache_entry = CacheEntry(
                    data=package_info,
                    timestamp=datetime.fromisoformat(data["timestamp"]),
                    version=data["version"],
                )
                return cache_entry, errors

            except json.JSONDecodeError:
                # Corrupted JSON file
                cache_file.unlink()
                errors.append(
                    FormattedError(
                        message=f"Corrupted cache file removed: {cache_key}",
                        suggestion="The cache file was corrupted and has been cleaned up automatically.",
                        severity=ErrorSeverity.WARNING,
                        error_code="cache_corruption_fixed",
                    )
                )
                return None, errors

        except Exception as e:
            formatted_error = ErrorFormatter.format_exception(
                e, {"cache_key": cache_key, "operation": "cache_retrieval"}
            )
            errors.append(formatted_error)
            return None, errors

    async def resolve_and_cache(
        self, package_name: str, version_constraint: str | None = None
    ) -> tuple[dict[str, Any], bool]:
        """
        Resolve package version and retrieve from cache or fetch fresh.

        Args:
            package_name: Name of the package
            version_constraint: Optional version constraint

        Returns:
            Tuple of (package_info_dict, from_cache_bool)
        """
        from .doc_fetcher import PyPIDocumentationFetcher
        from .version_resolver import VersionResolver

        # Initialize resolvers
        version_resolver = VersionResolver()

        try:
            # Step 1: Resolve to specific version
            resolved_version = await version_resolver.resolve_version(
                package_name, version_constraint
            )

            # Step 2: Check version-specific cache
            cache_key = version_resolver.generate_cache_key(
                package_name, resolved_version
            )
            cached_entry = await self.get(cache_key)

            if cached_entry:
                logger.debug(
                    "Cache hit for resolve_and_cache",
                    package=package_name,
                    version=resolved_version,
                )
                # Return as dictionary for compatibility
                return cached_entry.data.model_dump(), True
            else:
                logger.debug(
                    "Cache miss, fetching fresh",
                    package=package_name,
                    version=resolved_version,
                )

                # Step 3: Fetch fresh from PyPI
                async with PyPIDocumentationFetcher() as fetcher:
                    package_info = await fetcher.fetch_package_info(package_name)
                    await self.set(cache_key, package_info)

                # Return as dictionary for compatibility
                return package_info.model_dump(), False

        except Exception as e:
            logger.error(
                "Failed to resolve and cache package",
                package=package_name,
                constraint=version_constraint,
                error=str(e),
            )
            raise



"""Unit tests for documentation fetching functionality."""

import pytest
from httpx import Response

from src.autodoc_mcp.core.doc_fetcher import (
    DocFetchResult,
    PyPIDocumentationFetcher,
)
from src.autodoc_mcp.exceptions import NetworkError, PackageNotFoundError
from src.autodoc_mcp.models import PackageInfo


@pytest.fixture
def mock_config(mocker):
    """Mock configuration for tests."""
    config = mocker.Mock()
    config.pypi_base_url = "https://pypi.org/pypi"
    config.max_concurrent = 5
    config.max_documentation_size = 50000
    return config


@pytest.fixture
def sample_package_info():
    """Sample PackageInfo for tests."""
    return PackageInfo(
        name="requests",
        version="2.28.0",
        summary="Python HTTP for Humans.",
        description="Requests is a simple, yet elegant, HTTP library for Python.",
        author="Kenneth Reitz",
        license="Apache 2.0",
        home_page="https://requests.readthedocs.io",
        project_urls={
            "Documentation": "https://requests.readthedocs.io",
            "Source": "https://github.com/psf/requests",
        },
        classifiers=["Development Status :: 5 - Production/Stable"],
        keywords=["http", "requests", "web"],
    )


@pytest.fixture
def sample_pypi_response():
    """Sample PyPI API response."""
    return {
        "info": {
            "name": "requests",
            "version": "2.28.0",
            "summary": "Python HTTP for Humans.",
            "description": "Requests is a simple, yet elegant, HTTP library for Python.",
            "author": "Kenneth Reitz",
            "license": "Apache 2.0",
            "home_page": "https://requests.readthedocs.io",
            "project_urls": {
                "Documentation": "https://requests.readthedocs.io",
                "Source": "https://github.com/psf/requests",
            },
            "classifiers": ["Development Status :: 5 - Production/Stable"],
            "keywords": "http requests web",
        }
    }


class TestPyPIDocumentationFetcher:
    """Test PyPIDocumentationFetcher functionality."""

    @pytest.fixture
    def fetcher(self, mock_config, mocker):
        """Create fetcher instance with mocked config."""
        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            return PyPIDocumentationFetcher()

    @pytest.mark.asyncio
    async def test_context_manager_initialization(self, fetcher, mocker):
        """Test async context manager properly initializes client."""
        mock_client = mocker.AsyncMock()

        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.NetworkResilientClient",
            return_value=mock_client,
        ):
            async with fetcher:
                assert fetcher._resilient_client is mock_client
                mock_client.__aenter__.assert_called_once()

            mock_client.__aexit__.assert_called_once()

    @pytest.mark.asyncio
    async def test_fetch_package_info_success(
        self, fetcher, sample_pypi_response, sample_package_info, mocker
    ):
        """Test successful package info fetching."""
        # Mock HTTP response
        mock_response = mocker.Mock(spec=Response)
        mock_response.json.return_value = sample_pypi_response

        # Mock resilient client
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.return_value = mock_response
        fetcher._resilient_client = mock_client

        # Test fetch
        result = await fetcher.fetch_package_info("requests")

        # Verify result
        assert result.name == "requests"
        assert result.version == "2.28.0"
        assert result.summary == "Python HTTP for Humans."
        assert (
            result.description
            == "Requests is a simple, yet elegant, HTTP library for Python."
        )
        assert result.author == "Kenneth Reitz"

        # Verify API call
        mock_client.get_with_retry.assert_called_once_with(
            "https://pypi.org/pypi/requests/json",
            headers={"Accept": "application/json"},
        )

    @pytest.mark.asyncio
    async def test_fetch_package_info_package_not_found(self, fetcher, mocker):
        """Test handling of package not found errors."""
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.side_effect = PackageNotFoundError(
            "Package not found"
        )
        fetcher._resilient_client = mock_client

        with pytest.raises(PackageNotFoundError):
            await fetcher.fetch_package_info("nonexistent-package")

    @pytest.mark.asyncio
    async def test_fetch_package_info_network_error(self, fetcher, mocker):
        """Test handling of network errors."""
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.side_effect = NetworkError("Connection failed")
        fetcher._resilient_client = mock_client

        with pytest.raises(NetworkError):
            await fetcher.fetch_package_info("requests")

    @pytest.mark.asyncio
    async def test_fetch_package_info_client_not_initialized(self, fetcher):
        """Test error when client not initialized."""
        fetcher._resilient_client = None

        with pytest.raises(RuntimeError, match="Client not initialized"):
            await fetcher.fetch_package_info("requests")

    @pytest.mark.asyncio
    async def test_fetch_package_info_safe_success(
        self, fetcher, sample_pypi_response, sample_package_info, mocker
    ):
        """Test safe fetch with successful result."""
        # Mock HTTP response
        mock_response = mocker.Mock(spec=Response)
        mock_response.json.return_value = sample_pypi_response

        # Mock resilient client
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.return_value = mock_response
        fetcher._resilient_client = mock_client

        # Test safe fetch
        result = await fetcher.fetch_package_info_safe("requests")

        # Verify result structure
        assert isinstance(result, DocFetchResult)
        assert result.success is True
        assert result.package_info is not None
        assert result.package_info.name == "requests"
        assert len(result.errors) == 0
        assert len(result.warnings) == 0
        assert result.fetch_time > 0

    @pytest.mark.asyncio
    async def test_fetch_package_info_safe_package_not_found(self, fetcher, mocker):
        """Test safe fetch handles package not found gracefully."""
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.side_effect = PackageNotFoundError(
            "Package not found"
        )
        fetcher._resilient_client = mock_client

        result = await fetcher.fetch_package_info_safe("nonexistent-package")

        assert isinstance(result, DocFetchResult)
        assert result.success is False
        assert result.package_info is None
        assert len(result.errors) == 1
        assert result.errors[0].error_code is not None

    @pytest.mark.asyncio
    async def test_fetch_package_info_safe_network_error(self, fetcher, mocker):
        """Test safe fetch handles network errors gracefully."""
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.side_effect = NetworkError("Connection failed")
        fetcher._resilient_client = mock_client

        result = await fetcher.fetch_package_info_safe("requests")

        assert isinstance(result, DocFetchResult)
        assert result.success is False
        assert result.package_info is None
        assert len(result.errors) == 1
        assert result.errors[0].error_code is not None

    @pytest.mark.asyncio
    async def test_fetch_package_info_safe_unexpected_error(self, fetcher, mocker):
        """Test safe fetch handles unexpected errors gracefully."""
        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.side_effect = ValueError("Unexpected error")
        fetcher._resilient_client = mock_client

        result = await fetcher.fetch_package_info_safe("requests")

        assert isinstance(result, DocFetchResult)
        assert result.success is False
        assert result.package_info is None
        assert len(result.errors) == 1
        assert result.errors[0].error_code is not None

    @pytest.mark.asyncio
    async def test_fetch_multiple_packages_safe(
        self, fetcher, sample_pypi_response, mocker
    ):
        """Test fetching multiple packages safely."""
        # Mock successful responses for both packages
        mock_response = mocker.Mock(spec=Response)
        mock_response.json.return_value = sample_pypi_response

        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.return_value = mock_response
        fetcher._resilient_client = mock_client

        # Test batch fetch
        packages = ["requests", "httpx"]
        results = await fetcher.fetch_multiple_packages_safe(packages)

        # Verify results structure
        assert len(results) == 2
        assert "requests" in results
        assert "httpx" in results
        assert all(isinstance(r, DocFetchResult) for r in results.values())
        assert all(r.success for r in results.values())

    @pytest.mark.asyncio
    async def test_fetch_multiple_packages_safe_mixed_results(
        self, fetcher, sample_pypi_response, mocker
    ):
        """Test batch fetch with mixed success/failure results."""

        # Mock responses: success for requests, failure for nonexistent
        def mock_get_with_retry(url, **kwargs):
            if "requests" in url:
                mock_response = mocker.Mock(spec=Response)
                mock_response.json.return_value = sample_pypi_response
                return mock_response
            else:
                raise PackageNotFoundError("Package not found")

        mock_client = mocker.AsyncMock()
        mock_client.get_with_retry.side_effect = mock_get_with_retry
        fetcher._resilient_client = mock_client

        # Test batch fetch
        packages = ["requests", "nonexistent-package"]
        results = await fetcher.fetch_multiple_packages_safe(packages)

        # Verify mixed results
        assert len(results) == 2
        assert results["requests"].success is True
        assert results["nonexistent-package"].success is False
        assert len(results["nonexistent-package"].errors) == 1


class TestDocumentationFormatting:
    """Test documentation formatting functionality."""

    @pytest.fixture
    def fetcher(self, mock_config, mocker):
        """Create fetcher instance with mocked config."""
        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            return PyPIDocumentationFetcher()

    def test_format_documentation_basic(
        self, fetcher, sample_package_info, mock_config, mocker
    ):
        """Test basic documentation formatting."""
        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            formatted = fetcher.format_documentation(sample_package_info)

        # Check basic structure
        assert "# requests v2.28.0" in formatted
        assert "**Summary**: Python HTTP for Humans." in formatted
        assert "**Author**: Kenneth Reitz" in formatted
        assert "## Description" in formatted
        assert "## Links" in formatted
        assert "**Documentation**: https://requests.readthedocs.io" in formatted

    def test_format_documentation_with_query(
        self, fetcher, sample_package_info, mock_config, mocker
    ):
        """Test documentation formatting with query filtering."""
        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            formatted = fetcher.format_documentation(
                sample_package_info, query="http web"
            )

        # Query filtering returns sections that match the query terms
        # Since "http" is in the keywords, should return relevant content
        assert len(formatted) > 0
        # Should contain keywords section since "http" matches
        assert "http" in formatted.lower()

    def test_format_documentation_size_limits(
        self, fetcher, sample_package_info, mock_config, mocker
    ):
        """Test documentation size limiting and truncation."""
        # Create package with very long description
        long_description = "Very long description. " * 1000
        long_package = PackageInfo(
            name="test-package",
            version="1.0.0",
            summary="Test package",
            description=long_description,
            author="Test Author",
            license="MIT",
            home_page="https://example.com",
            project_urls={},
            classifiers=[],
            keywords=[],
        )

        # Set small max size for testing
        mock_config.max_documentation_size = 1000

        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            formatted = fetcher.format_documentation(long_package)

        # Should be truncated
        assert len(formatted) <= 1100  # Allow some buffer for truncation message
        assert "(truncated for performance)" in formatted

    def test_format_documentation_minimal_package_info(self, fetcher):
        """Test formatting with minimal package information."""
        minimal_package = PackageInfo(
            name="minimal-package",
            version="0.1.0",
            summary="",
            description="",
            author=None,
            license=None,
            home_page=None,
            project_urls={},
            classifiers=[],
            keywords=[],
        )

        formatted = fetcher.format_documentation(minimal_package)

        # Should handle missing fields gracefully
        assert "# minimal-package v0.1.0" in formatted
        assert len(formatted) > 0


class TestPyPIResponseParsing:
    """Test PyPI response parsing functionality."""

    @pytest.fixture
    def fetcher(self, mock_config, mocker):
        """Create fetcher instance with mocked config."""
        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            return PyPIDocumentationFetcher()

    def test_parse_pypi_response_valid(self, fetcher, sample_pypi_response):
        """Test parsing valid PyPI response."""
        package_info = fetcher._parse_pypi_response(sample_pypi_response, "requests")

        assert package_info.name == "requests"
        assert package_info.version == "2.28.0"
        assert package_info.summary == "Python HTTP for Humans."
        assert package_info.author == "Kenneth Reitz"
        assert len(package_info.keywords) == 3
        assert "http" in package_info.keywords

    def test_parse_pypi_response_missing_name(self, fetcher):
        """Test parsing response with missing name."""
        malformed_response = {
            "info": {
                "version": "1.0.0",
                "summary": "Test package",
            }
        }

        with pytest.raises(ValueError, match="Missing package name"):
            fetcher._parse_pypi_response(malformed_response, "test")

    def test_parse_pypi_response_missing_version(self, fetcher):
        """Test parsing response with missing version."""
        malformed_response = {
            "info": {
                "name": "test-package",
                "summary": "Test package",
            }
        }

        with pytest.raises(ValueError, match="Missing version"):
            fetcher._parse_pypi_response(malformed_response, "test-package")

    def test_parse_pypi_response_missing_info_key(self, fetcher):
        """Test parsing response with missing info key."""
        malformed_response = {
            "data": {
                "name": "test-package",
                "version": "1.0.0",
            }
        }

        with pytest.raises(ValueError, match="Malformed PyPI response"):
            fetcher._parse_pypi_response(malformed_response, "test-package")

    def test_parse_pypi_response_optional_fields(self, fetcher):
        """Test parsing response with only required fields."""
        minimal_response = {
            "info": {
                "name": "minimal-package",
                "version": "0.1.0",
            }
        }

        package_info = fetcher._parse_pypi_response(minimal_response, "minimal-package")

        assert package_info.name == "minimal-package"
        assert package_info.version == "0.1.0"
        assert package_info.summary == ""
        assert package_info.description == ""
        assert package_info.author is None
        assert package_info.keywords == []

    def test_parse_pypi_response_keywords_handling(self, fetcher):
        """Test proper keyword parsing from string."""
        response_with_keywords = {
            "info": {
                "name": "keyword-test",
                "version": "1.0.0",
                "keywords": "web http api rest",
            }
        }

        package_info = fetcher._parse_pypi_response(
            response_with_keywords, "keyword-test"
        )

        assert len(package_info.keywords) == 4
        assert "web" in package_info.keywords
        assert "http" in package_info.keywords
        assert "api" in package_info.keywords
        assert "rest" in package_info.keywords


class TestQueryFiltering:
    """Test query-based content filtering functionality."""

    @pytest.fixture
    def fetcher(self, mock_config, mocker):
        """Create fetcher instance with mocked config."""
        with mocker.patch(
            "src.autodoc_mcp.core.doc_fetcher.get_config", return_value=mock_config
        ):
            return PyPIDocumentationFetcher()

    def test_apply_query_filter_basic(self, fetcher):
        """Test basic query filtering functionality."""
        content = """# Package Name

## HTTP Client
This section talks about HTTP requests and web interactions.

## Database Support
This section covers database operations and SQL queries.

## Logging
This section covers logging and debugging features."""

        filtered = fetcher._apply_query_filter(content, "http web")

        # Should prioritize HTTP section
        assert "HTTP Client" in filtered
        assert "HTTP requests and web interactions" in filtered
        # May include other sections based on scoring

    def test_apply_query_filter_no_matches(self, fetcher):
        """Test query filtering with no matches."""
        content = """# Package Name

## Database Support
This section covers database operations.

## File Operations
This section covers file handling."""

        filtered = fetcher._apply_query_filter(content, "http networking")

        # Should return empty or minimal content when no matches
        assert len(filtered) >= 0  # May be empty or contain title

    def test_apply_query_filter_case_insensitive(self, fetcher):
        """Test case-insensitive query filtering."""
        content = """# Package Name

## HTTP Client
This section talks about HTTP requests."""

        filtered = fetcher._apply_query_filter(content, "HTTP")

        assert "HTTP Client" in filtered
        assert "HTTP requests" in filtered



"""User-friendly error message formatting with actionable suggestions."""

import re
from dataclasses import dataclass
from enum import Enum
from re import Match
from typing import Any

from ..exceptions import (
    NetworkError,
    PackageNotFoundError,
    ProjectParsingError,
)


class ErrorSeverity(Enum):
    """Error severity levels for UI/UX purposes."""

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


@dataclass
class FormattedError:
    """User-friendly error with context and suggestions."""

    message: str
    suggestion: str | None = None
    severity: ErrorSeverity = ErrorSeverity.ERROR
    error_code: str | None = None
    context: dict[str, Any] | None = None
    recoverable: bool = True


class ErrorFormatter:
    """Converts technical errors to user-friendly messages."""

    # Common dependency parsing error patterns
    DEPENDENCY_PATTERNS = [
        (r"Invalid package name format: '([^']+)'", "invalid_package_name"),
        (r"Unclosed bracket in extras: '([^']+)'", "malformed_extras"),
        (r"Empty dependency string", "empty_dependency"),
        (r"Malformed extras in '([^']+)'", "malformed_extras"),
        (r"Invalid version specifier.*'([^']+)'", "invalid_version"),
    ]

    @classmethod
    def format_dependency_parsing_errors(
        cls, failed_deps: list[dict[str, Any]]
    ) -> list[FormattedError]:
        """Format dependency parsing errors with suggestions."""
        formatted_errors = []

        for failure in failed_deps:
            dep_str = failure.get("dependency_string", "unknown")
            error_msg = failure.get("error", "")
            source = failure.get("source", "unknown")

            formatted = cls._format_single_dependency_error(dep_str, error_msg, source)
            formatted_errors.append(formatted)

        return formatted_errors

    @classmethod
    def _format_single_dependency_error(
        cls, dep_str: str, error_msg: str, source: str
    ) -> FormattedError:
        """Format a single dependency parsing error."""

        # Check against known patterns
        for pattern, error_type in cls.DEPENDENCY_PATTERNS:
            match = re.search(pattern, error_msg)
            if match:
                return cls._create_dependency_error_by_type(
                    error_type, dep_str, source, match
                )

        # Fallback for unknown errors
        return FormattedError(
            message=f"Could not parse dependency '{dep_str}' from {source} section",
            suggestion=f"Check the syntax of '{dep_str}'. Common formats: package>=1.0.0, package[extra1,extra2]>=1.0.0",
            severity=ErrorSeverity.WARNING,
            error_code="dependency_parse_unknown",
            recoverable=True,
        )

    @classmethod
    def _create_dependency_error_by_type(
        cls, error_type: str, dep_str: str, source: str, match: Match[str]
    ) -> FormattedError:
        """Create specific error messages based on error type."""

        if error_type == "invalid_package_name":
            package_name = match.group(1) if match.groups() else dep_str
            return FormattedError(
                message=f"Invalid package name '{package_name}' in {source} dependencies",
                suggestion=f"Package names can only contain letters, numbers, hyphens, periods, and underscores. Try: {cls._suggest_package_name_fix(package_name)}",
                severity=ErrorSeverity.WARNING,
                error_code="invalid_package_name",
                context={"original_name": package_name, "source": source},
            )

        elif error_type == "malformed_extras":
            return FormattedError(
                message=f"Malformed extras syntax in '{dep_str}' from {source} section",
                suggestion=f"Use square brackets for extras: {cls._suggest_extras_fix(dep_str)}",
                severity=ErrorSeverity.WARNING,
                error_code="malformed_extras",
                context={"original_dep": dep_str, "source": source},
            )

        elif error_type == "invalid_version":
            return FormattedError(
                message=f"Invalid version specifier in '{dep_str}' from {source} section",
                suggestion=f"Use valid version operators: {cls._suggest_version_fix(dep_str)}",
                severity=ErrorSeverity.WARNING,
                error_code="invalid_version",
                context={"original_dep": dep_str, "source": source},
            )

        elif error_type == "empty_dependency":
            return FormattedError(
                message=f"Empty dependency entry found in {source} section",
                suggestion="Remove empty strings from your dependencies list",
                severity=ErrorSeverity.WARNING,
                error_code="empty_dependency",
                context={"source": source},
            )

        # Fallback
        return FormattedError(
            message=f"Could not parse '{dep_str}' from {source} section",
            suggestion="Check dependency syntax and try again",
            severity=ErrorSeverity.WARNING,
            error_code="dependency_parse_unknown",
        )

    @classmethod
    def _suggest_package_name_fix(cls, name: str) -> str:
        """Suggest a fixed package name."""
        # Replace common problematic characters
        fixed = re.sub(r"[^A-Za-z0-9._-]", "-", name)
        fixed = re.sub(r"--+", "-", fixed)  # Replace multiple hyphens
        fixed = fixed.strip("-")  # Remove leading/trailing hyphens
        return fixed or "package-name"

    @classmethod
    def _suggest_extras_fix(cls, dep_str: str) -> str:
        """Suggest a fixed extras syntax."""
        # Extract package name before any brackets or version specs
        package_match = re.match(r"^([A-Za-z0-9._-]+)", dep_str)
        if package_match:
            return f"{package_match.group(1)}[extra1,extra2]>=1.0.0"
        return "package[extra1,extra2]>=1.0.0"

    @classmethod
    def _suggest_version_fix(cls, dep_str: str) -> str:
        """Suggest a fixed version syntax."""
        # Extract package name
        package_match = re.match(r"^([A-Za-z0-9._-]+)", dep_str)
        if package_match:
            return f"{package_match.group(1)}>=1.0.0"
        return "package>=1.0.0"

    @classmethod
    def format_network_errors(
        cls, errors: list[dict[str, Any]]
    ) -> list[FormattedError]:
        """Format network-related errors with suggestions."""
        formatted_errors = []

        for error in errors:
            error_type = error.get("type", "")
            package = error.get("package", "unknown")
            message = error.get("message", "")

            if error_type == "PackageNotFoundError":
                suggestions = error.get("suggestions", [])
                suggestion_text = ""
                if suggestions:
                    suggestion_text = f" Did you mean: {', '.join(suggestions[:3])}?"

                formatted_errors.append(
                    FormattedError(
                        message=f"Package '{package}' not found on PyPI",
                        suggestion=f"Check the package name spelling.{suggestion_text}",
                        severity=ErrorSeverity.ERROR,
                        error_code="package_not_found",
                        context={"package": package, "suggestions": suggestions},
                        recoverable=True,
                    )
                )

            elif error_type == "NetworkError":
                retry_suggested = error.get("retry_suggested", False)
                suggestion = "Check your internet connection and try again."
                if retry_suggested:
                    suggestion += " The request will be retried automatically."

                formatted_errors.append(
                    FormattedError(
                        message=f"Network error while fetching '{package}': {message}",
                        suggestion=suggestion,
                        severity=ErrorSeverity.ERROR,
                        error_code="network_error",
                        context={"package": package},
                        recoverable=True,
                    )
                )

            else:
                # Generic error formatting
                formatted_errors.append(
                    FormattedError(
                        message=f"Error processing '{package}': {message}",
                        suggestion="Review the error details and try again.",
                        severity=ErrorSeverity.ERROR,
                        error_code="generic_error",
                        context={"package": package, "type": error_type},
                    )
                )

        return formatted_errors

    @classmethod
    def format_exception(
        cls, exc: Exception, context: dict[str, Any] | None = None
    ) -> FormattedError:
        """Format any exception into a user-friendly error."""

        if isinstance(exc, ProjectParsingError):
            return FormattedError(
                message=f"Could not parse project file: {exc}",
                suggestion="Check your pyproject.toml file for syntax errors",
                severity=ErrorSeverity.CRITICAL,
                error_code="project_parse_error",
                context={"file_path": str(exc.file_path) if exc.file_path else None},
                recoverable=False,
            )

        elif isinstance(exc, PackageNotFoundError):
            package = context.get("package", "unknown") if context else "unknown"
            return FormattedError(
                message=f"Package '{package}' not found",
                suggestion="Verify the package name and check if it exists on PyPI",
                severity=ErrorSeverity.ERROR,
                error_code="package_not_found",
                context=context,
                recoverable=True,
            )

        elif isinstance(exc, NetworkError):
            return FormattedError(
                message=f"Network error: {exc}",
                suggestion="Check your internet connection. The operation will be retried automatically.",
                severity=ErrorSeverity.ERROR,
                error_code="network_error",
                context=context,
                recoverable=True,
            )

        else:
            return FormattedError(
                message=f"Unexpected error: {exc}",
                suggestion="This may be a system issue. Please try again or report if the problem persists.",
                severity=ErrorSeverity.CRITICAL,
                error_code="unexpected_error",
                context={"exception_type": type(exc).__name__},
                recoverable=False,
            )


class ResponseFormatter:
    """Format complete responses with errors, warnings, and suggestions."""

    @classmethod
    def format_scan_response(cls, result: Any) -> dict[str, Any]:
        """Format scan results with user-friendly error messages."""

        # Format dependency parsing errors
        formatted_dependency_errors = ErrorFormatter.format_dependency_parsing_errors(
            result.failed_deps
        )

        # Create suggestions summary
        suggestions = [
            err.suggestion for err in formatted_dependency_errors if err.suggestion
        ]

        return {
            "success": len(result.dependencies) > 0,
            "partial_success": result.partial_success,
            "project_name": result.project_name,
            "project_path": str(result.project_path),
            # Successful results
            "dependencies": [dep.model_dump() for dep in result.dependencies],
            "dependency_count": result.successful_deps,
            # Error information (user-friendly)
            "errors": [
                {
                    "message": err.message,
                    "suggestion": err.suggestion,
                    "severity": err.severity.value,
                    "code": err.error_code,
                    "recoverable": err.recoverable,
                }
                for err in formatted_dependency_errors
            ],
            # Legacy warning format (for backwards compatibility)
            "warnings": result.warnings,
            # Actionable suggestions summary
            "suggestions": suggestions,
            "scan_timestamp": result.scan_timestamp.isoformat(),
        }



"""Tests for enhanced configuration validation."""

import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from autodoc_mcp.config import AutoDocsConfig


class TestAutoDocsConfigValidation:
    """Test configuration validation."""

    def test_default_config_creation(self):
        """Test creating config with defaults."""
        config = AutoDocsConfig()

        assert config.max_concurrent == 10
        assert config.request_timeout == 30
        assert config.log_level == "INFO"
        assert config.pypi_base_url == "https://pypi.org/pypi"
        assert config.environment == "development"
        assert config.debug_mode is False

    def test_field_validation_constraints(self):
        """Test field validation constraints."""
        # Test max_concurrent constraints
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(max_concurrent=0)  # Below minimum
        assert "Input should be greater than or equal to 1" in str(exc_info.value)

        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(max_concurrent=101)  # Above maximum
        assert "Input should be less than or equal to 100" in str(exc_info.value)

        # Test request_timeout constraints
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(request_timeout=4)  # Below minimum
        assert "Input should be greater than or equal to 5" in str(exc_info.value)

        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(request_timeout=301)  # Above maximum
        assert "Input should be less than or equal to 300" in str(exc_info.value)

    def test_log_level_validation(self):
        """Test log level validation."""
        # Valid log levels
        for level in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            config = AutoDocsConfig(log_level=level)
            assert config.log_level == level

        # Case insensitive
        config = AutoDocsConfig(log_level="debug")
        assert config.log_level == "DEBUG"

        # Invalid log level
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(log_level="INVALID")
        assert "Invalid log level: INVALID" in str(exc_info.value)

    def test_pypi_url_validation(self):
        """Test PyPI URL validation."""
        # Valid URLs
        valid_urls = [
            "https://pypi.org/pypi",
            "https://test.pypi.org/pypi",
            "http://localhost:8080/pypi",
        ]
        for url in valid_urls:
            config = AutoDocsConfig(pypi_base_url=url)
            assert config.pypi_base_url == url

        # Invalid URLs
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(pypi_base_url="invalid-url")
        assert "Invalid PyPI URL" in str(exc_info.value)

        # Untrusted domain
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(pypi_base_url="https://evil.com/pypi")
        assert "Untrusted PyPI domain" in str(exc_info.value)

    def test_environment_validation(self):
        """Test environment validation."""
        # Valid environments
        for env in ["development", "staging", "production"]:
            config = AutoDocsConfig(environment=env)
            assert config.environment == env

        # Invalid environment
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(environment="invalid")
        assert "Invalid environment: invalid" in str(exc_info.value)

    def test_retry_delay_validation(self):
        """Test retry delay validation."""
        # Valid configuration
        config = AutoDocsConfig(base_retry_delay=1.0, max_retry_delay=30.0)
        assert config.base_retry_delay == 1.0
        assert config.max_retry_delay == 30.0

        # Invalid: max <= base
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(base_retry_delay=5.0, max_retry_delay=5.0)
        assert "max_retry_delay must be greater than base_retry_delay" in str(
            exc_info.value
        )

        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(base_retry_delay=10.0, max_retry_delay=5.0)
        assert "max_retry_delay must be greater than base_retry_delay" in str(
            exc_info.value
        )

    def test_context_settings_validation(self):
        """Test context settings validation."""
        # Valid settings
        config = AutoDocsConfig(
            max_dependency_context=5,
            max_context_tokens=15000,
        )
        assert config.max_dependency_context == 5
        assert config.max_context_tokens == 15000

        # Invalid max_dependency_context
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(max_dependency_context=0)
        assert "Input should be greater than or equal to 1" in str(exc_info.value)

        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(max_dependency_context=21)
        assert "Input should be less than or equal to 20" in str(exc_info.value)

        # Invalid max_context_tokens
        with pytest.raises(ValidationError) as exc_info:
            AutoDocsConfig(max_context_tokens=500)  # Below minimum
        assert "Input should be greater than or equal to 1000" in str(exc_info.value)

    def test_cache_dir_validation_and_creation(self):
        """Test cache directory validation and creation."""
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_path = Path(temp_dir) / "test_cache"

            # Should create the directory
            config = AutoDocsConfig(cache_dir=cache_path)
            assert config.cache_dir == cache_path
            assert cache_path.exists()
            assert cache_path.is_dir()

    def test_cache_dir_write_permission_check(self):
        """Test cache directory write permission check."""
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_path = Path(temp_dir) / "readonly_cache"
            cache_path.mkdir()

            # Make directory read-only (this might not work on all systems)
            try:
                cache_path.chmod(0o444)

                # This should fail due to write permission check
                with pytest.raises(ValidationError) as exc_info:
                    AutoDocsConfig(cache_dir=cache_path)
                assert "not writable" in str(exc_info.value)

            finally:
                # Restore permissions for cleanup
                cache_path.chmod(0o755)

    def test_from_env_basic(self):
        """Test loading configuration from environment."""
        env_vars = {
            "AUTODOCS_MAX_CONCURRENT": "20",
            "AUTODOCS_REQUEST_TIMEOUT": "45",
            "AUTODOCS_LOG_LEVEL": "DEBUG",
            "AUTODOCS_ENVIRONMENT": "staging",
            "AUTODOCS_DEBUG": "true",
        }

        with patch.dict(os.environ, env_vars, clear=False):
            config = AutoDocsConfig.from_env()

            assert config.max_concurrent == 20
            assert config.request_timeout == 45
            assert config.log_level == "DEBUG"
            assert config.environment == "staging"
            assert config.debug_mode is True

    def test_from_env_with_cache_dir(self):
        """Test loading cache directory from environment."""
        with tempfile.TemporaryDirectory() as temp_dir:
            env_vars = {
                "AUTODOCS_CACHE_DIR": temp_dir,
            }

            with patch.dict(os.environ, env_vars, clear=False):
                config = AutoDocsConfig.from_env()
                assert config.cache_dir == Path(temp_dir)

    def test_from_env_invalid_values(self):
        """Test handling invalid environment variable values."""
        env_vars = {
            "AUTODOCS_MAX_CONCURRENT": "invalid",
        }

        with patch.dict(os.environ, env_vars, clear=False):
            with pytest.raises(ValueError) as exc_info:
                AutoDocsConfig.from_env()
            assert "Configuration validation failed" in str(exc_info.value)

    def test_from_env_boolean_conversion(self):
        """Test boolean conversion from environment."""
        test_cases = [
            ("true", True),
            ("True", True),
            ("TRUE", True),
            ("false", False),
            ("False", False),
            ("anything_else", False),
        ]

        for env_value, expected in test_cases:
            env_vars = {"AUTODOCS_DEBUG": env_value}

            with patch.dict(os.environ, env_vars, clear=False):
                config = AutoDocsConfig.from_env()
                assert config.debug_mode is expected

    def test_from_env_numeric_conversions(self):
        """Test numeric conversions from environment."""
        env_vars = {
            "AUTODOCS_MAX_RETRY_ATTEMPTS": "5",
            "AUTODOCS_BASE_RETRY_DELAY": "2.5",
            "AUTODOCS_CIRCUIT_BREAKER_TIMEOUT": "120.0",
        }

        with patch.dict(os.environ, env_vars, clear=False):
            config = AutoDocsConfig.from_env()
            assert config.max_retry_attempts == 5
            assert config.base_retry_delay == 2.5
            assert config.circuit_breaker_timeout == 120.0

    def test_validate_production_readiness_development(self):
        """Test production readiness validation in development."""
        config = AutoDocsConfig(
            environment="development",
            debug_mode=True,
            log_level="DEBUG",
        )

        issues = config.validate_production_readiness()
        assert len(issues) == 0  # No issues in development

    def test_validate_production_readiness_production_good(self):
        """Test production readiness validation with good production config."""
        config = AutoDocsConfig(
            environment="production",
            debug_mode=False,
            log_level="INFO",
            pypi_base_url="https://pypi.org/pypi",
            max_concurrent=30,
        )

        issues = config.validate_production_readiness()
        assert len(issues) == 0

    def test_validate_production_readiness_production_issues(self):
        """Test production readiness validation with issues."""
        config = AutoDocsConfig(
            environment="production",
            debug_mode=True,  # Issue 1
            log_level="DEBUG",  # Issue 2
            pypi_base_url="http://localhost:8080/pypi",  # Issue 3 (not HTTPS)
            max_concurrent=75,  # Issue 4 (high concurrency)
        )

        issues = config.validate_production_readiness()
        assert len(issues) == 4

        issue_text = " ".join(issues)
        assert "debug mode" in issue_text.lower()
        assert "debug logging" in issue_text.lower()
        assert "https required" in issue_text.lower()
        assert "high concurrency" in issue_text.lower()

    def test_comprehensive_configuration_validation(self):
        """Test comprehensive configuration with all settings."""
        config_data = {
            "cache_dir": Path("/tmp/test_cache"),
            "max_concurrent": 15,
            "request_timeout": 60,
            "log_level": "WARNING",
            "pypi_base_url": "https://test.pypi.org/pypi",
            "max_cached_versions_per_package": 5,
            "cache_cleanup_days": 30,
            "max_dependency_context": 10,
            "max_context_tokens": 20000,
            "max_retry_attempts": 5,
            "base_retry_delay": 2.0,
            "max_retry_delay": 60.0,
            "circuit_breaker_threshold": 10,
            "circuit_breaker_timeout": 120.0,
            "rate_limit_requests_per_minute": 120,
            "max_documentation_size": 100000,
            "environment": "staging",
            "debug_mode": True,
        }

        # Should not raise any validation errors
        config = AutoDocsConfig(**config_data)

        # Verify all values are set correctly
        for key, value in config_data.items():
            assert getattr(config, key) == value



"""Observability and metrics collection for AutoDocs MCP Server."""

import time
import uuid
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import Any

import structlog

logger = structlog.get_logger(__name__)


@dataclass
class RequestMetrics:
    """Track request performance metrics."""

    request_id: str
    operation: str
    start_time: float = field(default_factory=time.time)
    end_time: float | None = None
    success: bool = False
    error_type: str | None = None
    cache_hit: bool = False
    package_name: str | None = None
    dependency_count: int = 0

    @property
    def duration_ms(self) -> float:
        """Calculate request duration in milliseconds."""
        end = self.end_time or time.time()
        return (end - self.start_time) * 1000

    def to_dict(self) -> dict[str, Any]:
        """Convert metrics to dictionary for logging."""
        return {
            "request_id": self.request_id,
            "operation": self.operation,
            "duration_ms": round(self.duration_ms, 2),
            "success": self.success,
            "error_type": self.error_type,
            "cache_hit": self.cache_hit,
            "package_name": self.package_name,
            "dependency_count": self.dependency_count,
            "timestamp": self.start_time,
        }


class MetricsCollector:
    """Collect and aggregate performance metrics."""

    def __init__(self) -> None:
        self.active_requests: dict[str, RequestMetrics] = {}
        self.completed_requests: list[RequestMetrics] = []
        self.max_completed = 1000  # Keep last 1000 requests

    def start_request(self, request_id: str, operation: str) -> RequestMetrics:
        """Start tracking a request."""
        metrics = RequestMetrics(request_id=request_id, operation=operation)
        self.active_requests[request_id] = metrics
        return metrics

    def finish_request(
        self,
        request_id: str,
        success: bool = True,
        error_type: str | None = None,
        cache_hit: bool = False,
        package_name: str | None = None,
        dependency_count: int = 0,
    ) -> None:
        """Finish tracking a request."""
        if request_id in self.active_requests:
            metrics = self.active_requests.pop(request_id)
            metrics.end_time = time.time()
            metrics.success = success
            metrics.error_type = error_type
            metrics.cache_hit = cache_hit
            metrics.package_name = package_name
            metrics.dependency_count = dependency_count

            # Log the completed request
            logger.info("Request completed", **metrics.to_dict())

            # Store for aggregation (ring buffer)
            self.completed_requests.append(metrics)
            if len(self.completed_requests) > self.max_completed:
                self.completed_requests.pop(0)

    def get_stats(self) -> dict[str, Any]:
        """Get aggregated performance statistics."""
        if not self.completed_requests:
            return {
                "total_requests": 0,
                "active_requests": len(self.active_requests),
                "cache_hit_rate": 0.0,
                "success_rate": 0.0,
                "operations": {},
                "response_times": {"avg_ms": 0.0},
            }

        total = len(self.completed_requests)
        successful = sum(1 for r in self.completed_requests if r.success)
        cache_hits = sum(1 for r in self.completed_requests if r.cache_hit)

        durations = [r.duration_ms for r in self.completed_requests]

        # Calculate percentiles
        sorted_durations = sorted(durations)
        p50_idx = max(0, int(0.5 * len(sorted_durations)) - 1)
        p95_idx = max(0, int(0.95 * len(sorted_durations)) - 1)
        p99_idx = max(0, int(0.99 * len(sorted_durations)) - 1)

        # Operation breakdown
        operations = {}
        for r in self.completed_requests:
            op = r.operation
            if op not in operations:
                operations[op] = {
                    "count": 0,
                    "success_count": 0,
                    "avg_duration_ms": 0.0,
                    "cache_hits": 0,
                }
            operations[op]["count"] += 1
            if r.success:
                operations[op]["success_count"] += 1
            if r.cache_hit:
                operations[op]["cache_hits"] += 1

        # Calculate averages for operations
        for op_name, op_stats in operations.items():
            op_requests = [r for r in self.completed_requests if r.operation == op_name]
            if op_requests:
                op_durations = [r.duration_ms for r in op_requests]
                op_stats["avg_duration_ms"] = round(
                    sum(op_durations) / len(op_durations), 2
                )

        return {
            "total_requests": total,
            "success_rate": round(successful / total * 100, 2),
            "cache_hit_rate": round(cache_hits / total * 100, 2),
            "active_requests": len(self.active_requests),
            "operations": operations,
            "response_times": {
                "avg_ms": round(sum(durations) / len(durations), 2),
                "p50_ms": round(sorted_durations[p50_idx], 2),
                "p95_ms": round(sorted_durations[p95_idx], 2),
                "p99_ms": round(sorted_durations[p99_idx], 2),
                "min_ms": round(min(durations), 2),
                "max_ms": round(max(durations), 2),
            },
            "timestamp": time.time(),
        }

    def get_health_metrics(self) -> dict[str, Any]:
        """Get metrics relevant for health checks."""
        if not self.completed_requests:
            return {
                "active_requests": len(self.active_requests),
                "total_completed": 0,
                "recent_errors": 0,
            }

        # Look at last 100 requests for health indicators
        recent_requests = self.completed_requests[-100:]
        recent_errors = sum(1 for r in recent_requests if not r.success)

        return {
            "active_requests": len(self.active_requests),
            "total_completed": len(self.completed_requests),
            "recent_errors": recent_errors,
            "recent_requests": len(recent_requests),
            "error_rate_recent": round(recent_errors / len(recent_requests) * 100, 2)
            if recent_requests
            else 0.0,
        }


# Global metrics collector instance
metrics_collector = MetricsCollector()


@asynccontextmanager
async def track_request(
    operation: str, request_id: str | None = None
) -> AsyncGenerator[RequestMetrics, None]:
    """Context manager to track request metrics."""
    if request_id is None:
        request_id = str(uuid.uuid4())[:8]

    metrics = metrics_collector.start_request(request_id, operation)

    try:
        yield metrics
        # Don't automatically finish here - let the caller set specific details
    except Exception as e:
        metrics_collector.finish_request(
            request_id, success=False, error_type=type(e).__name__
        )
        raise


def get_metrics_collector() -> MetricsCollector:
    """Get the global metrics collector instance."""
    return metrics_collector


def setup_production_logging() -> None:
    """Configure structured logging for production."""
    import logging
    import sys

    processors: list[Any] = [
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.JSONRenderer(),  # JSON for production
    ]

    structlog.configure(
        processors=processors,
        wrapper_class=structlog.make_filtering_bound_logger(logging.INFO),
        logger_factory=structlog.WriteLoggerFactory(file=sys.stderr),
        cache_logger_on_first_use=True,
    )


def setup_development_logging() -> None:
    """Configure structured logging for development."""
    import logging
    import sys

    processors: list[Any] = [
        structlog.processors.TimeStamper(fmt="ISO"),
        structlog.processors.add_log_level,
        structlog.dev.ConsoleRenderer(),  # Pretty console output for dev
    ]

    structlog.configure(
        processors=processors,
        wrapper_class=structlog.make_filtering_bound_logger(logging.DEBUG),
        logger_factory=structlog.WriteLoggerFactory(file=sys.stderr),
        cache_logger_on_first_use=True,
    )



"""Concurrent context fetcher for high-performance dependency documentation."""

import asyncio
import time
from typing import Any

from structlog import get_logger

from ..config import get_config
from .cache_manager import CacheManagerInterface
from .context_formatter import (
    ContextDocumentationFormatter,
    ContextWindowManager,
    DocumentationContext,
    PackageDocumentation,
)
from .dependency_resolver import DependencyResolver
from .network_client import NetworkResilientClient

logger = get_logger(__name__)


class ConcurrentContextFetcher:
    """High-performance concurrent fetcher for documentation context."""

    def __init__(
        self,
        cache_manager: CacheManagerInterface,
        dependency_resolver: DependencyResolver,
        formatter: ContextDocumentationFormatter,
        network_client: NetworkResilientClient | None = None,
    ):
        self.cache_manager = cache_manager
        self.dependency_resolver = dependency_resolver
        self.formatter = formatter
        self.network_client = network_client
        self.config = get_config()

    async def fetch_package_context(
        self,
        package_name: str,
        version_constraint: str | None = None,
        include_dependencies: bool = True,
        context_scope: str = "smart",
        max_dependencies: int | None = None,
        max_tokens: int | None = None,
    ) -> tuple[DocumentationContext, dict[str, Any]]:
        """
        Fetch comprehensive documentation context for a package.

        Args:
            package_name: Primary package name
            version_constraint: Version constraint for primary package
            include_dependencies: Whether to include dependency context
            context_scope: Scope of context ("primary_only", "runtime", "smart")
            max_dependencies: Maximum dependencies to include
            max_tokens: Maximum token budget for context

        Returns:
            Tuple of (DocumentationContext, performance_metrics)
        """
        start_time = time.time()
        performance_metrics = {
            "primary_fetch_time": 0.0,
            "dependency_resolution_time": 0.0,
            "dependency_fetch_time": 0.0,
            "total_time": 0.0,
            "cache_hits": 0,
            "cache_misses": 0,
            "concurrent_fetches": 0,
            "failed_fetches": 0,
        }

        try:
            logger.info(
                "Starting context fetch",
                package=package_name,
                include_deps=include_dependencies,
                scope=context_scope,
            )

            # Step 1: Fetch primary package documentation
            primary_start = time.time()
            primary_info, from_cache = await self.cache_manager.resolve_and_cache(
                package_name, version_constraint
            )

            if from_cache:
                performance_metrics["cache_hits"] += 1
            else:
                performance_metrics["cache_misses"] += 1

            primary_docs = self.formatter.format_primary_package(
                primary_info, from_cache
            )
            performance_metrics["primary_fetch_time"] = time.time() - primary_start

            # Initialize context with primary package
            context = DocumentationContext(
                primary_package=primary_docs, context_scope="primary_only"
            )

            # If dependencies not requested, return early
            if not include_dependencies:
                performance_metrics["total_time"] = time.time() - start_time
                return context, performance_metrics

            # Step 2: Resolve dependency context
            resolution_start = time.time()
            max_deps = max_dependencies or (
                self.config.max_dependency_context if context_scope == "smart" else 5
            )
            max_context_tokens = max_tokens or self.config.max_context_tokens

            dependency_names = (
                await self.dependency_resolver.resolve_context_dependencies(
                    package_name=primary_info["name"],
                    version=primary_info["version"],
                    max_dependencies=max_deps,
                    max_tokens=max_context_tokens,
                )
            )

            performance_metrics["dependency_resolution_time"] = (
                time.time() - resolution_start
            )

            if not dependency_names:
                logger.info("No dependencies to fetch", package=package_name)
                performance_metrics["total_time"] = time.time() - start_time
                return context, performance_metrics

            # Step 3: Fetch dependency documentation concurrently
            dep_fetch_start = time.time()
            dependency_results = await self._fetch_dependencies_concurrently(
                dependency_names, primary_info["name"], performance_metrics
            )

            performance_metrics["dependency_fetch_time"] = time.time() - dep_fetch_start
            performance_metrics["concurrent_fetches"] = len(dependency_names)

            # Step 4: Process successful results
            successful_deps = []
            failed_deps = []

            for result in dependency_results:
                if isinstance(result, Exception):
                    failed_deps.append(str(result))
                    performance_metrics["failed_fetches"] += 1
                else:
                    successful_deps.append(result)

            # Update context with dependencies
            context.runtime_dependencies = successful_deps
            context.total_packages = 1 + len(successful_deps)

            if successful_deps:
                context.context_scope = f"with_runtime ({len(successful_deps)} deps)"

            if failed_deps:
                context.truncated_packages.extend(
                    [f"Failed: {dep}" for dep in failed_deps]
                )

            # Step 5: Apply token window management
            context = ContextWindowManager.fit_to_window(context, max_context_tokens)

            performance_metrics["total_time"] = time.time() - start_time

            logger.info(
                "Context fetch complete",
                package=package_name,
                total_packages=context.total_packages,
                total_time=performance_metrics["total_time"],
                tokens=context.token_estimate,
            )

            return context, performance_metrics

        except Exception as e:
            performance_metrics["total_time"] = time.time() - start_time
            logger.error(
                "Context fetch failed",
                package=package_name,
                error=str(e),
                total_time=performance_metrics["total_time"],
            )

            # Return minimal context with primary package only
            if "primary_docs" in locals():
                context = DocumentationContext(
                    primary_package=primary_docs,
                    context_scope="primary_only (deps_failed)",
                )
                return context, performance_metrics

            raise

    async def _fetch_dependencies_concurrently(
        self,
        dependency_names: list[str],
        primary_package_name: str,
        performance_metrics: dict[str, Any],
    ) -> list[Any]:
        """Fetch multiple dependencies concurrently with proper error handling."""

        if not dependency_names:
            return []

        # Create concurrent tasks
        tasks = []
        semaphore = asyncio.Semaphore(5)  # Limit concurrent requests

        for dep_name in dependency_names:
            task = asyncio.create_task(
                self._fetch_single_dependency_with_semaphore(
                    semaphore, dep_name, primary_package_name, performance_metrics
                )
            )
            tasks.append(task)

        # Wait for all tasks with timeout
        try:
            results = await asyncio.wait_for(
                asyncio.gather(*tasks, return_exceptions=True),
                timeout=15.0,  # Don't wait too long for dependencies
            )
            return results

        except TimeoutError:
            logger.warning(
                "Dependency fetching timed out",
                dependencies=dependency_names,
                timeout=15.0,
            )

            # Cancel remaining tasks
            for task in tasks:
                if not task.done():
                    task.cancel()

            # Return whatever we got before timeout
            completed_results: list[Any] = []
            for task in tasks:
                if task.done() and not task.cancelled():
                    try:
                        result = task.result()
                        completed_results.append(result)
                    except Exception as e:
                        completed_results.append(e)
                else:
                    completed_results.append(Exception("Timeout fetching dependency"))

            return completed_results

    async def _fetch_single_dependency_with_semaphore(
        self,
        semaphore: asyncio.Semaphore,
        dep_name: str,
        primary_package_name: str,
        performance_metrics: dict[str, Any],
    ) -> PackageDocumentation:
        """Fetch a single dependency with semaphore control."""

        async with semaphore:
            return await self._fetch_single_dependency(
                dep_name, primary_package_name, performance_metrics
            )

    async def _fetch_single_dependency(
        self,
        dep_name: str,
        primary_package_name: str,
        performance_metrics: dict[str, Any],
    ) -> PackageDocumentation:
        """Fetch documentation for a single dependency."""

        try:
            logger.debug("Fetching dependency", dependency=dep_name)

            # Try to resolve and cache the dependency
            dep_info, from_cache = await self.cache_manager.resolve_and_cache(dep_name)

            if from_cache:
                performance_metrics["cache_hits"] += 1
            else:
                performance_metrics["cache_misses"] += 1

            # Format as dependency documentation
            dep_docs = self.formatter.format_dependency_package(
                dep_info, primary_package_name
            )

            logger.debug(
                "Successfully fetched dependency",
                dependency=dep_name,
                from_cache=from_cache,
            )

            return dep_docs

        except Exception as e:
            logger.warning(
                "Failed to fetch dependency", dependency=dep_name, error=str(e)
            )

            # Return minimal documentation for failed fetches
            return PackageDocumentation(
                name=dep_name,
                version="unknown",
                relationship="runtime_dependency",
                why_included=f"Required by {primary_package_name} (fetch failed)",
                dependency_level=1,
                summary=f"Documentation fetch failed: {str(e)[:100]}...",
            )


async def create_context_fetcher(
    cache_manager: CacheManagerInterface,
) -> ConcurrentContextFetcher:
    """Create a concurrent context fetcher with all dependencies."""

    # Initialize network client
    network_client = NetworkResilientClient()
    await network_client.__aenter__()

    # Initialize dependency resolver
    dependency_resolver = DependencyResolver(network_client)

    # Initialize formatter
    formatter = ContextDocumentationFormatter()

    return ConcurrentContextFetcher(
        cache_manager=cache_manager,
        dependency_resolver=dependency_resolver,
        formatter=formatter,
        network_client=network_client,
    )



"""Security validation and sanitization functions."""

import re
from pathlib import Path
from urllib.parse import urlparse

# Trusted PyPI domains for URL validation
TRUSTED_PYPI_DOMAINS: set[str] = {
    "pypi.org",
    "test.pypi.org",
    "pypi.python.org",  # Legacy support
}


def validate_pypi_url(url: str) -> str:
    """Validate PyPI URL against allowlist of trusted domains.

    Args:
        url: The PyPI URL to validate

    Returns:
        The validated URL

    Raises:
        ValueError: If URL is invalid or uses untrusted domain
    """
    if not isinstance(url, str) or not url.strip():
        raise ValueError("URL must be a non-empty string")

    try:
        parsed = urlparse(url.strip())

        if not parsed.scheme or parsed.scheme not in ("http", "https"):
            raise ValueError(
                f"Invalid URL scheme: {parsed.scheme}. Must be http or https"
            )

        if not parsed.netloc:
            raise ValueError("URL must have a valid domain")

        # Check against trusted domains
        domain = parsed.netloc.lower()
        if domain not in TRUSTED_PYPI_DOMAINS and not domain.startswith("localhost"):
            raise ValueError(
                f"Untrusted PyPI domain: {domain}. Allowed domains: {TRUSTED_PYPI_DOMAINS}"
            )

        # Ensure HTTPS for production (except localhost)
        if parsed.scheme == "http" and not domain.startswith("localhost"):
            raise ValueError(
                "HTTP URLs not allowed for production PyPI domains. Use HTTPS"
            )

        return url.strip()

    except ValueError:
        raise
    except Exception as e:
        raise ValueError(f"Invalid PyPI URL '{url}': {e}") from e


def sanitize_cache_key(cache_key: str) -> str:
    """Sanitize cache key to prevent path traversal attacks.

    Args:
        cache_key: The cache key to sanitize

    Returns:
        The sanitized cache key

    Raises:
        ValueError: If cache key is invalid after sanitization
    """
    if not isinstance(cache_key, str):
        raise ValueError("Cache key must be a string")

    # Remove dangerous characters and path components
    sanitized = re.sub(r'[<>:"/\\|?*\x00-\x1f]', "_", cache_key)

    # Collapse multiple dots to prevent ../ traversal
    sanitized = re.sub(r"\.\.+", ".", sanitized)

    # Remove leading/trailing dots and spaces
    sanitized = sanitized.strip(". ")

    # Ensure reasonable length
    if len(sanitized) > 200:
        sanitized = sanitized[:200]

    # Validate final result
    if not sanitized or sanitized in {"", ".", ".."}:
        raise ValueError(f"Invalid cache key after sanitization: '{cache_key}'")

    # Check for reserved Windows filenames
    reserved_names = {
        "CON",
        "PRN",
        "AUX",
        "NUL",
        "COM1",
        "COM2",
        "COM3",
        "COM4",
        "COM5",
        "COM6",
        "COM7",
        "COM8",
        "COM9",
        "LPT1",
        "LPT2",
        "LPT3",
        "LPT4",
        "LPT5",
        "LPT6",
        "LPT7",
        "LPT8",
        "LPT9",
    }
    if sanitized.upper() in reserved_names:
        raise ValueError(
            f"Invalid cache key after sanitization (reserved name): '{cache_key}'"
        )

    return sanitized


class InputValidator:
    """Centralized input validation for all user inputs."""

    @staticmethod
    def validate_package_name(name: str) -> str:
        """Validate PyPI package name format.

        Args:
            name: Package name to validate

        Returns:
            Normalized package name (lowercase)

        Raises:
            ValueError: If package name is invalid
        """
        if not isinstance(name, str) or not name:
            raise ValueError("Package name must be a non-empty string")

        # PyPI naming rules: https://packaging.python.org/en/latest/specifications/name-normalization/
        if not re.match(r"^[A-Za-z0-9]([A-Za-z0-9._-]*[A-Za-z0-9])?$", name):
            raise ValueError(
                f"Invalid package name format: '{name}'. Must start and end with alphanumeric characters"
            )

        if len(name) > 214:  # PyPI limit
            raise ValueError(f"Package name too long: {len(name)} > 214 characters")

        return name.lower()  # Normalize case

    @staticmethod
    def validate_version_constraint(constraint: str) -> str:
        """Validate version constraint syntax.

        Args:
            constraint: Version constraint to validate

        Returns:
            The validated constraint

        Raises:
            ValueError: If version constraint is invalid
        """
        if not isinstance(constraint, str):
            raise ValueError("Version constraint must be a string")

        if not constraint.strip():
            raise ValueError("Version constraint cannot be empty")

        try:
            from packaging.specifiers import SpecifierSet

            SpecifierSet(constraint)
            return constraint.strip()
        except Exception as e:
            raise ValueError(f"Invalid version constraint '{constraint}': {e}") from e

    @staticmethod
    def validate_project_path(path: str) -> Path:
        """Validate and resolve project path.

        Args:
            path: Project path to validate

        Returns:
            Resolved Path object

        Raises:
            ValueError: If path is invalid
        """
        if not isinstance(path, str) or not path.strip():
            raise ValueError("Project path must be a non-empty string")

        try:
            resolved = Path(path).resolve()
            if not resolved.exists():
                raise ValueError(f"Path does not exist: {path}")
            if not resolved.is_dir():
                raise ValueError(f"Path is not a directory: {path}")
            return resolved
        except (OSError, RuntimeError) as e:
            raise ValueError(f"Invalid project path '{path}': {e}") from e



"""Unit tests for network client functionality."""

import httpx
import pytest

from src.autodoc_mcp.core.network_client import (
    BasicNetworkClient,
    NetworkResilientClient,
)
from src.autodoc_mcp.exceptions import NetworkError, PackageNotFoundError


class TestBasicNetworkClientInitialization:
    """Test network client initialization and lifecycle."""

    def test_init_creates_none_client(self):
        """Test that initialization sets client to None."""
        client = BasicNetworkClient()
        assert client._client is None

    @pytest.mark.asyncio
    async def test_aenter_creates_client(self, mocker):
        """Test async context manager entry creates httpx client."""
        mock_httpx_client = mocker.patch("httpx.AsyncClient")
        mock_instance = mocker.AsyncMock()
        mock_httpx_client.return_value = mock_instance

        client = BasicNetworkClient()
        result = await client.__aenter__()

        assert result is client
        assert client._client is mock_instance
        mock_httpx_client.assert_called_once_with(
            timeout=httpx.Timeout(30.0),
            follow_redirects=True,
            headers={"User-Agent": "AutoDocs-MCP/1.0"},
        )

    @pytest.mark.asyncio
    async def test_aexit_closes_client(self, mocker):
        """Test async context manager exit closes httpx client."""
        mock_client = mocker.AsyncMock()

        client = BasicNetworkClient()
        client._client = mock_client

        await client.__aexit__(None, None, None)

        mock_client.aclose.assert_called_once()

    @pytest.mark.asyncio
    async def test_aexit_with_none_client(self):
        """Test async context manager exit with None client."""
        client = BasicNetworkClient()
        client._client = None

        # Should not raise exception
        await client.__aexit__(None, None, None)

    @pytest.mark.asyncio
    async def test_aexit_handles_exception_context(self, mocker):
        """Test async context manager exit with exception context."""
        mock_client = mocker.AsyncMock()

        client = BasicNetworkClient()
        client._client = mock_client

        # Pass exception context
        await client.__aexit__(ValueError, ValueError("test"), None)

        mock_client.aclose.assert_called_once()


class TestNetworkResilientClientAlias:
    """Test NetworkResilientClient alias."""

    def test_alias_points_to_basic_client(self):
        """Test that NetworkResilientClient is an alias for BasicNetworkClient."""
        assert NetworkResilientClient is BasicNetworkClient

    def test_alias_creates_same_instance(self):
        """Test that alias creates the same type of instance."""
        client1 = BasicNetworkClient()
        client2 = NetworkResilientClient()

        assert type(client1) is type(client2)
        assert isinstance(client2, BasicNetworkClient)


class TestGetWithRetryClientValidation:
    """Test client validation in get_with_retry."""

    @pytest.mark.asyncio
    async def test_get_with_retry_no_client_raises_error(self):
        """Test get_with_retry raises NetworkError when client is None."""
        client = BasicNetworkClient()
        # Don't initialize client
        assert client._client is None

        with pytest.raises(NetworkError, match="HTTP client not initialized"):
            await client.get_with_retry("https://example.com")


class TestGetWithRetrySuccessScenarios:
    """Test successful request scenarios."""

    @pytest.mark.asyncio
    async def test_get_with_retry_success_first_attempt(self, mocker):
        """Test successful request on first attempt."""
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        mock_response.raise_for_status = mocker.Mock()

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(return_value=mock_response)

        client = BasicNetworkClient()
        client._client = mock_client

        result = await client.get_with_retry("https://example.com", param="value")

        assert result is mock_response
        mock_client.get.assert_called_once_with("https://example.com", param="value")
        mock_response.raise_for_status.assert_called_once()

    @pytest.mark.asyncio
    async def test_get_with_retry_success_after_failures(self, mocker):
        """Test successful request after some failures."""
        # First two attempts fail, third succeeds
        mock_timeout_error = httpx.TimeoutException("Timeout")
        mock_success_response = mocker.Mock()
        mock_success_response.status_code = 200
        mock_success_response.raise_for_status = mocker.Mock()

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(
            side_effect=[mock_timeout_error, mock_timeout_error, mock_success_response]
        )

        client = BasicNetworkClient()
        client._client = mock_client

        # Mock asyncio.sleep to avoid actual delays in tests
        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        result = await client.get_with_retry("https://example.com")

        assert result is mock_success_response
        assert mock_client.get.call_count == 3
        # Should have slept twice (after first and second failures)
        assert mock_sleep.call_count == 2
        mock_sleep.assert_any_call(1)  # 2^(1-1) = 1
        mock_sleep.assert_any_call(2)  # 2^(2-1) = 2


class TestGetWithRetry404Handling:
    """Test 404 handling scenarios."""

    @pytest.mark.asyncio
    async def test_get_with_retry_404_status_code(self, mocker):
        """Test 404 status code raises PackageNotFoundError."""
        mock_response = mocker.Mock()
        mock_response.status_code = 404

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(return_value=mock_response)

        client = BasicNetworkClient()
        client._client = mock_client

        with pytest.raises(
            PackageNotFoundError, match="Resource not found: https://example.com"
        ):
            await client.get_with_retry("https://example.com")

    @pytest.mark.asyncio
    async def test_get_with_retry_404_http_status_error(self, mocker):
        """Test 404 HTTPStatusError raises PackageNotFoundError."""
        mock_response = mocker.Mock()
        mock_response.status_code = 404

        mock_http_error = httpx.HTTPStatusError(
            message="Not Found", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        with pytest.raises(
            PackageNotFoundError, match="Resource not found: https://example.com"
        ):
            await client.get_with_retry("https://example.com")


class TestGetWithRetryTimeoutHandling:
    """Test timeout handling and backoff logic."""

    @pytest.mark.asyncio
    async def test_get_with_retry_timeout_all_attempts_fail(self, mocker):
        """Test timeout on all attempts raises NetworkError."""
        mock_timeout_error = httpx.TimeoutException("Request timeout")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_timeout_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError, match="Request timeout after 3 attempts"):
            await client.get_with_retry("https://example.com")

        # Should attempt 3 times
        assert mock_client.get.call_count == 3
        # Should sleep twice (after first two failures)
        assert mock_sleep.call_count == 2
        mock_sleep.assert_any_call(1)  # 2^(1-1) = 1
        mock_sleep.assert_any_call(2)  # 2^(2-1) = 2

    @pytest.mark.asyncio
    async def test_get_with_retry_timeout_exponential_backoff(self, mocker):
        """Test timeout uses exponential backoff correctly."""
        mock_timeout_error = httpx.TimeoutException("Request timeout")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_timeout_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError):
            await client.get_with_retry("https://example.com")

        # Verify exponential backoff: 2^0=1, 2^1=2
        sleep_calls = [call[0][0] for call in mock_sleep.call_args_list]
        assert sleep_calls == [1, 2]


class TestGetWithRetryHTTPStatusErrorHandling:
    """Test HTTP status error handling for different status codes."""

    @pytest.mark.asyncio
    async def test_get_with_retry_4xx_client_error_no_retry(self, mocker):
        """Test 4xx client errors (except 408, 429) don't retry."""
        mock_response = mocker.Mock()
        mock_response.status_code = 400
        mock_response.text = "Bad Request Error Message"

        mock_http_error = httpx.HTTPStatusError(
            message="Bad Request", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        with pytest.raises(NetworkError, match="HTTP 400: Bad Request Error Message"):
            await client.get_with_retry("https://example.com")

        # Should only attempt once (no retry for 4xx except 408, 429)
        assert mock_client.get.call_count == 1

    @pytest.mark.asyncio
    async def test_get_with_retry_408_timeout_retries(self, mocker):
        """Test 408 Request Timeout retries with backoff."""
        mock_response = mocker.Mock()
        mock_response.status_code = 408
        mock_response.text = "Request Timeout"

        mock_http_error = httpx.HTTPStatusError(
            message="Request Timeout", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError, match="HTTP error after 3 attempts"):
            await client.get_with_retry("https://example.com")

        # Should retry 3 times for 408
        assert mock_client.get.call_count == 3
        assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_get_with_retry_429_rate_limit_retries(self, mocker):
        """Test 429 Rate Limit retries with backoff."""
        mock_response = mocker.Mock()
        mock_response.status_code = 429
        mock_response.text = "Too Many Requests"

        mock_http_error = httpx.HTTPStatusError(
            message="Too Many Requests", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError, match="HTTP error after 3 attempts"):
            await client.get_with_retry("https://example.com")

        # Should retry 3 times for 429
        assert mock_client.get.call_count == 3
        assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_get_with_retry_5xx_server_error_retries(self, mocker):
        """Test 5xx server errors retry with backoff."""
        mock_response = mocker.Mock()
        mock_response.status_code = 500
        mock_response.text = "Internal Server Error"

        mock_http_error = httpx.HTTPStatusError(
            message="Internal Server Error",
            request=mocker.Mock(),
            response=mock_response,
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError, match="HTTP error after 3 attempts"):
            await client.get_with_retry("https://example.com")

        # Should retry 3 times for 5xx errors
        assert mock_client.get.call_count == 3
        assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_get_with_retry_403_forbidden_no_retry(self, mocker):
        """Test 403 Forbidden doesn't retry."""
        mock_response = mocker.Mock()
        mock_response.status_code = 403
        mock_response.text = "Forbidden Access"

        mock_http_error = httpx.HTTPStatusError(
            message="Forbidden", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        with pytest.raises(NetworkError, match="HTTP 403: Forbidden Access"):
            await client.get_with_retry("https://example.com")

        # Should only attempt once
        assert mock_client.get.call_count == 1

    @pytest.mark.asyncio
    async def test_get_with_retry_response_text_truncation(self, mocker):
        """Test response text is truncated to 200 chars in error message."""
        long_text = "x" * 300  # 300 character error message

        mock_response = mocker.Mock()
        mock_response.status_code = 400
        mock_response.text = long_text

        mock_http_error = httpx.HTTPStatusError(
            message="Bad Request", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        with pytest.raises(NetworkError) as exc_info:
            await client.get_with_retry("https://example.com")

        # Error message should contain only first 200 characters
        error_message = str(exc_info.value)
        assert "x" * 200 in error_message
        assert len([char for char in error_message if char == "x"]) == 200


class TestGetWithRetryRequestErrorHandling:
    """Test request error handling scenarios."""

    @pytest.mark.asyncio
    async def test_get_with_retry_request_error_all_attempts_fail(self, mocker):
        """Test RequestError on all attempts raises NetworkError."""
        mock_request_error = httpx.RequestError("Connection failed")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_request_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(
            NetworkError, match="Network error after 3 attempts: Connection failed"
        ):
            await client.get_with_retry("https://example.com")

        # Should attempt 3 times
        assert mock_client.get.call_count == 3
        # Should sleep twice (after first two failures)
        assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_get_with_retry_request_error_success_after_retry(self, mocker):
        """Test RequestError followed by success."""
        mock_request_error = httpx.RequestError("Connection failed")
        mock_success_response = mocker.Mock()
        mock_success_response.status_code = 200
        mock_success_response.raise_for_status = mocker.Mock()

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(
            side_effect=[mock_request_error, mock_success_response]
        )

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        result = await client.get_with_retry("https://example.com")

        assert result is mock_success_response
        assert mock_client.get.call_count == 2
        assert mock_sleep.call_count == 1  # One sleep after first failure
        mock_sleep.assert_called_with(1)  # 2^(1-1) = 1


class TestGetWithRetryMixedErrorScenarios:
    """Test mixed error scenarios and edge cases."""

    @pytest.mark.asyncio
    async def test_get_with_retry_mixed_errors_with_success(self, mocker):
        """Test mix of different error types followed by success."""
        mock_timeout_error = httpx.TimeoutException("Timeout")

        mock_response_429 = mocker.Mock()
        mock_response_429.status_code = 429
        mock_response_429.text = "Rate Limited"
        mock_http_error_429 = httpx.HTTPStatusError(
            message="Too Many Requests",
            request=mocker.Mock(),
            response=mock_response_429,
        )

        mock_success_response = mocker.Mock()
        mock_success_response.status_code = 200
        mock_success_response.raise_for_status = mocker.Mock()

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(
            side_effect=[mock_timeout_error, mock_http_error_429, mock_success_response]
        )

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        result = await client.get_with_retry("https://example.com")

        assert result is mock_success_response
        assert mock_client.get.call_count == 3
        assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_get_with_retry_max_attempts_boundary(self, mocker):
        """Test that exactly 3 attempts are made before giving up."""
        mock_request_error = httpx.RequestError("Connection failed")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_request_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError):
            await client.get_with_retry("https://example.com")

        # Verify exactly 3 attempts
        assert mock_client.get.call_count == 3
        # Verify exactly 2 sleep calls (between attempts)
        assert mock_sleep.call_count == 2

    @pytest.mark.asyncio
    async def test_get_with_retry_max_attempts_boundary_verification(self, mocker):
        """Test that exactly max_attempts are made before giving up."""
        # Test ensures the loop logic works correctly
        mock_request_error = httpx.RequestError("Connection failed")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_request_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_sleep = mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError, match="Network error after 3 attempts"):
            await client.get_with_retry("https://example.com")

        # Verify exactly 3 attempts as expected
        assert mock_client.get.call_count == 3
        # Verify exactly 2 sleep calls (between attempts)
        assert mock_sleep.call_count == 2


class TestGetWithRetryLoggingVerification:
    """Test that proper logging occurs during retry attempts."""

    @pytest.mark.asyncio
    async def test_get_with_retry_debug_logging_on_success(self, mocker):
        """Test debug logging occurs on successful requests."""
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        mock_response.raise_for_status = mocker.Mock()

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(return_value=mock_response)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_logger = mocker.patch("src.autodoc_mcp.core.network_client.logger")

        await client.get_with_retry("https://example.com")

        mock_logger.debug.assert_called_with(
            "Making HTTP request", url="https://example.com", attempt=1
        )

    @pytest.mark.asyncio
    async def test_get_with_retry_warning_logging_on_timeout(self, mocker):
        """Test warning logging occurs on timeout."""
        mock_timeout_error = httpx.TimeoutException("Timeout")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_timeout_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_logger = mocker.patch("src.autodoc_mcp.core.network_client.logger")
        mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError):
            await client.get_with_retry("https://example.com")

        # Should log warning for each timeout attempt
        assert mock_logger.warning.call_count == 3
        mock_logger.warning.assert_any_call(
            "Request timeout", url="https://example.com", attempt=1
        )
        mock_logger.warning.assert_any_call(
            "Request timeout", url="https://example.com", attempt=2
        )
        mock_logger.warning.assert_any_call(
            "Request timeout", url="https://example.com", attempt=3
        )

    @pytest.mark.asyncio
    async def test_get_with_retry_warning_logging_on_http_error(self, mocker):
        """Test warning logging occurs on retryable HTTP errors."""
        mock_response = mocker.Mock()
        mock_response.status_code = 500
        mock_response.text = "Server Error"

        mock_http_error = httpx.HTTPStatusError(
            message="Server Error", request=mocker.Mock(), response=mock_response
        )

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_http_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_logger = mocker.patch("src.autodoc_mcp.core.network_client.logger")
        mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError):
            await client.get_with_retry("https://example.com")

        # Should log warning for each HTTP error attempt
        assert mock_logger.warning.call_count == 3
        mock_logger.warning.assert_any_call(
            "HTTP error",
            url="https://example.com",
            status=500,
            attempt=1,
        )

    @pytest.mark.asyncio
    async def test_get_with_retry_warning_logging_on_request_error(self, mocker):
        """Test warning logging occurs on request errors."""
        mock_request_error = httpx.RequestError("Connection failed")

        mock_client = mocker.AsyncMock()
        mock_client.get = mocker.AsyncMock(side_effect=mock_request_error)

        client = BasicNetworkClient()
        client._client = mock_client

        mock_logger = mocker.patch("src.autodoc_mcp.core.network_client.logger")
        mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError):
            await client.get_with_retry("https://example.com")

        # Should log warning for each request error attempt
        assert mock_logger.warning.call_count == 3
        mock_logger.warning.assert_any_call(
            "Network error",
            url="https://example.com",
            error="Connection failed",
            attempt=1,
        )


class TestAsyncContextManagerIntegration:
    """Test full async context manager integration scenarios."""

    @pytest.mark.asyncio
    async def test_full_context_manager_workflow(self, mocker):
        """Test complete async context manager workflow."""
        mock_response = mocker.Mock()
        mock_response.status_code = 200
        mock_response.raise_for_status = mocker.Mock()

        mock_httpx_client = mocker.patch("httpx.AsyncClient")
        mock_client_instance = mocker.AsyncMock()
        mock_client_instance.get = mocker.AsyncMock(return_value=mock_response)
        mock_httpx_client.return_value = mock_client_instance

        async with BasicNetworkClient() as client:
            result = await client.get_with_retry("https://example.com")

        assert result is mock_response
        mock_client_instance.aclose.assert_called_once()

    @pytest.mark.asyncio
    async def test_context_manager_with_exception(self, mocker):
        """Test context manager cleanup when exception occurs."""
        mock_httpx_client = mocker.patch("httpx.AsyncClient")
        mock_client_instance = mocker.AsyncMock()
        mock_client_instance.get = mocker.AsyncMock(
            side_effect=httpx.RequestError("Connection failed")
        )
        mock_httpx_client.return_value = mock_client_instance

        mocker.patch("asyncio.sleep", new_callable=mocker.AsyncMock)

        with pytest.raises(NetworkError, match="Network error after 3 attempts"):
            async with BasicNetworkClient() as client:
                await client.get_with_retry("https://example.com")

        # Should still close even when exception occurs
        mock_client_instance.aclose.assert_called_once()



"""Tests for security validation functions."""

import tempfile
from pathlib import Path

import pytest

from src.autodoc_mcp.security import (
    InputValidator,
    sanitize_cache_key,
    validate_pypi_url,
)


class TestValidatePyPIURL:
    """Test PyPI URL validation."""

    def test_valid_https_urls(self):
        """Test valid HTTPS URLs."""
        valid_urls = [
            "https://pypi.org/pypi",
            "https://test.pypi.org/pypi",
            "https://pypi.python.org/pypi",
        ]

        for url in valid_urls:
            result = validate_pypi_url(url)
            assert result == url

    def test_valid_localhost_urls(self):
        """Test localhost URLs are allowed."""
        localhost_urls = [
            "http://localhost:8000/pypi",
            "https://localhost/pypi",
            "http://localhost/test",
        ]

        for url in localhost_urls:
            result = validate_pypi_url(url)
            assert result == url

    def test_invalid_empty_url(self):
        """Test empty URLs are rejected."""
        with pytest.raises(ValueError, match="URL must be a non-empty string"):
            validate_pypi_url("")

        with pytest.raises(ValueError, match="URL must be a non-empty string"):
            validate_pypi_url("   ")

    def test_invalid_none_url(self):
        """Test None URLs are rejected."""
        with pytest.raises(ValueError, match="URL must be a non-empty string"):
            validate_pypi_url(None)

    def test_invalid_scheme(self):
        """Test invalid URL schemes."""
        invalid_schemes = [
            "ftp://pypi.org/pypi",
            "file:///etc/passwd",
            "javascript:alert('xss')",
        ]

        for url in invalid_schemes:
            with pytest.raises(ValueError, match="Invalid URL scheme"):
                validate_pypi_url(url)

    def test_untrusted_domain(self):
        """Test untrusted domains are rejected."""
        untrusted_urls = [
            "https://evil.com/pypi",
            "https://malicious.org/pypi",
            "http://attacker.net/pypi",
        ]

        for url in untrusted_urls:
            with pytest.raises(ValueError, match="Untrusted PyPI domain"):
                validate_pypi_url(url)

    def test_http_production_domains_rejected(self):
        """Test HTTP URLs for production domains are rejected."""
        http_urls = ["http://pypi.org/pypi", "http://test.pypi.org/pypi"]

        for url in http_urls:
            with pytest.raises(
                ValueError, match="HTTP URLs not allowed for production"
            ):
                validate_pypi_url(url)

    def test_malformed_urls(self):
        """Test malformed URLs are rejected."""
        malformed_urls = [
            "not-a-url",
            "://missing-scheme",
            "https://",
            "https://pypi.org with spaces",
        ]

        for url in malformed_urls:
            with pytest.raises(ValueError):
                validate_pypi_url(url)

    def test_whitespace_trimming(self):
        """Test URLs with whitespace are trimmed."""
        url_with_whitespace = "  https://pypi.org/pypi  "
        result = validate_pypi_url(url_with_whitespace)
        assert result == "https://pypi.org/pypi"


class TestSanitizeCacheKey:
    """Test cache key sanitization."""

    def test_valid_cache_key(self):
        """Test valid cache keys pass through."""
        valid_keys = ["requests-2.28.2", "package_name-1.0.0", "my-package-v1.2.3"]

        for key in valid_keys:
            result = sanitize_cache_key(key)
            assert result == key

    def test_dangerous_characters_replaced(self):
        """Test dangerous characters are replaced."""
        dangerous_key = 'package<>:"/\\|?*\x01name'
        result = sanitize_cache_key(dangerous_key)
        assert (
            result == "package__________name"
        )  # 10 underscores for 10 dangerous chars

    def test_path_traversal_prevention(self):
        """Test path traversal attempts are prevented."""
        traversal_attempts = [
            "../../../etc/passwd",
            "..\\..\\windows\\system32",
            "package..name",
            "...///package",
        ]

        for key in traversal_attempts:
            result = sanitize_cache_key(key)
            assert ".." not in result

    def test_leading_trailing_dots_stripped(self):
        """Test leading/trailing dots are stripped."""
        dotted_keys = [".package", "package.", "..package..", "   .package.   "]

        expected = ["package", "package", "package", "package"]

        for key, expected_result in zip(dotted_keys, expected, strict=False):
            result = sanitize_cache_key(key)
            assert result == expected_result

    def test_long_keys_truncated(self):
        """Test long keys are truncated."""
        long_key = "a" * 300
        result = sanitize_cache_key(long_key)
        assert len(result) == 200

    def test_reserved_names_rejected(self):
        """Test Windows reserved names are rejected."""
        reserved_names = ["CON", "PRN", "AUX", "NUL", "COM1", "LPT1"]

        for name in reserved_names:
            with pytest.raises(ValueError, match="reserved name"):
                sanitize_cache_key(name)

    def test_empty_keys_rejected(self):
        """Test empty keys are rejected."""
        empty_keys = ["", "   ", ".", "..", "..."]

        for key in empty_keys:
            with pytest.raises(ValueError, match="Invalid cache key"):
                sanitize_cache_key(key)

    def test_non_string_input_rejected(self):
        """Test non-string inputs are rejected."""
        with pytest.raises(ValueError, match="Cache key must be a string"):
            sanitize_cache_key(123)

        with pytest.raises(ValueError, match="Cache key must be a string"):
            sanitize_cache_key(None)


class TestInputValidator:
    """Test input validation."""

    def test_validate_package_name_valid(self):
        """Test valid package names."""
        valid_names = [
            "requests",
            "Django",
            "package-name",
            "package_name",
            "package.name",
            "a",
            "a1",
            "1a",
        ]

        for name in valid_names:
            result = InputValidator.validate_package_name(name)
            assert result == name.lower()

    def test_validate_package_name_invalid(self):
        """Test invalid package names."""
        invalid_names = [
            "",
            "-invalid",
            "invalid-",
            "_invalid",
            "invalid_",
            ".invalid",
            "invalid.",
            "invalid@name",
            "invalid name",
            "a" * 250,  # Too long
        ]

        for name in invalid_names:
            with pytest.raises(ValueError):
                InputValidator.validate_package_name(name)

    def test_validate_package_name_non_string(self):
        """Test non-string package names are rejected."""
        with pytest.raises(ValueError, match="Package name must be a non-empty string"):
            InputValidator.validate_package_name(None)

        with pytest.raises(ValueError, match="Package name must be a non-empty string"):
            InputValidator.validate_package_name(123)

    def test_validate_version_constraint_valid(self):
        """Test valid version constraints."""
        valid_constraints = [">=1.0.0", "~=2.1.0", ">=1.0,<2.0", "==1.2.3", "!=1.0.0"]

        for constraint in valid_constraints:
            result = InputValidator.validate_version_constraint(constraint)
            assert result == constraint

    def test_validate_version_constraint_invalid(self):
        """Test invalid version constraints."""
        invalid_constraints = [
            "",
            "   ",
            "invalid",
            ">==1.0.0",  # Invalid operator
            ">=1.0.0 invalid",
        ]

        for constraint in invalid_constraints:
            with pytest.raises(ValueError):
                InputValidator.validate_version_constraint(constraint)

    def test_validate_version_constraint_non_string(self):
        """Test non-string version constraints are rejected."""
        with pytest.raises(ValueError, match="Version constraint must be a string"):
            InputValidator.validate_version_constraint(None)

    def test_validate_project_path_valid(self):
        """Test valid project paths."""
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)

            # Create a test file in the directory
            test_file = temp_path / "test.txt"
            test_file.write_text("test")

            result = InputValidator.validate_project_path(str(temp_path))
            assert result == temp_path.resolve()

    def test_validate_project_path_nonexistent(self):
        """Test nonexistent paths are rejected."""
        with pytest.raises(ValueError, match="Path does not exist"):
            InputValidator.validate_project_path("/nonexistent/path")

    def test_validate_project_path_not_directory(self):
        """Test file paths (not directories) are rejected."""
        with (
            tempfile.NamedTemporaryFile() as temp_file,
            pytest.raises(ValueError, match="Path is not a directory"),
        ):
            InputValidator.validate_project_path(temp_file.name)

    def test_validate_project_path_empty(self):
        """Test empty paths are rejected."""
        with pytest.raises(ValueError, match="Project path must be a non-empty string"):
            InputValidator.validate_project_path("")

        with pytest.raises(ValueError, match="Project path must be a non-empty string"):
            InputValidator.validate_project_path(None)
